# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.

from typing import Dict, List, Tuple
import numpy as np
from cigsegy.cpp._CXX_SEGY import Pysegy
from cigsegy import utils
import warnings


def read_header(fname: str, type, n=0, printstr=True):
    """
    Read binary or trace header

    Parameters
    ----------
    segy : str
        input segy file
    type: str
        can be one of ['bh', 'th'],
            'bt' means binary header, 'th' means trace header,
    n : int
        trace number when type is 'th'
    printstr : bool
        print header information, if False, return a dict of header's infomation

    Returns
    -------
    Dict or None
    """
    if type == 'bh':
        try:
            segy = Pysegy(fname)
            arr = segy.get_binary_header()
        except:
            warnings.warn("The SEG-Y file is broken, try to read binary header by numpy")
            arr = np.fromfile(fname, dtype=np.uint8, count=400, offset=3200)

        out, hstring = utils.parse_bheader(arr)
    elif type == 'th':
        segy = Pysegy(fname)
        arr = segy.get_trace_header(n)
        out, hstring = utils.parse_theader(arr)
    else:
        raise ValueError("type must be one of ['bh', 'th']")

    if printstr:
        print('\n'.join(hstring))
        return
    return out


def get_metaInfo(
    segyname: str,
    iline: int = None,
    xline: int = None,
    offset: int = None,
    istep: int = None,
    xstep: int = None,
    ostep: int = None,
    xloc: int = None,
    yloc: int = None,
    *,
    is4d: bool = None,
    apply_scalar: bool = False,
) -> Dict:
    """
    get metainfo dict of `segyname` file

    Parameters
    ----------
    segyname : str
        input segy file
    iline : int
        iline location in trace header
    iline : int
        iline location in trace header
    istep : int
        iline step, 2 means iline is like 100, 102, 104, ...
    xstep : int
        xline step
    xloc : int
        x (real world) value location in trace header
    yloc : int
        y (real world) value location in trace header
    apply_scalar : bool
        apply scalar to 'i-interval' and 'x-interval'

    Returns
    -------
    Dict
        Dict of meta information 
    """
    if isinstance(segyname, Pysegy):
        segy = segyname
    else:
        segy = Pysegy(str(segyname))

    [iline, xline, offset, istep, xstep, ostep, xloc, yloc, _is4d] = utils.guess(segy, iline, xline, offset, istep, xstep, ostep, xloc, yloc) # yapf: disable
    if is4d is None:
        is4d = _is4d
    # print(iline, xline, offset, istep, xstep, ostep, xloc, yloc, is4d)
    segy.setLocations(iline, xline, offset)
    segy.setSteps(istep, xstep, ostep)
    segy.setXYLocations(xloc, yloc)
    ndim = 4 if is4d else 3
    segy.set_segy_type(ndim)
    segy.scan()
    keys = segy.get_keylocs()
    meta = segy.get_metainfo()
    meta = {**keys, **meta}

    meta = utils.post_process_meta(segy, meta, apply_scalar)

    if not isinstance(segyname, Pysegy):
        segy.close()

    return meta


def get_lineInfo(
    fname: str,
    iline: int = None,
    xline: int = None,
    offset: int = None,
    istep: int = None,
    xstep: int = None,
    ostep: int = None,
    xloc: int = None,
    yloc: int = None,
    mode: str = 'raw',
):
    """
    mode can be one of ['raw', 'geom']
    """
    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))

    [iline, xline, offset, istep, xstep, ostep, xloc, yloc, is4d] = utils.guess(segy, iline, xline, offset, istep, xstep, ostep, xloc, yloc) # yapf: disable
    segy.setLocations(iline, xline, offset)
    segy.setSteps(istep, xstep, ostep)
    segy.setXYLocations(xloc, yloc)
    ndim = 4 if is4d else 3
    segy.set_segy_type(ndim)
    segy.scan()
    lineinfo = segy.get_lineInfo()
    ndim = segy.ndim

    out = None
    if mode == 'raw':
        out = lineinfo
    elif mode == 'geom':
        if ndim == 4:
            raise NotImplementedError("4D geometry is not supported yet")
        lineinfo = lineinfo[~np.any(lineinfo == -1, axis=1)]
        N = lineinfo.shape[0]
        out = np.zeros((N * 2 + 1, 4), dtype=np.int32)
        out[:N, :2] = lineinfo[:, :2]
        for i in range(N):
            idx = lineinfo[i, 3]
            out[i, 2:] = [segy.coordx(idx), segy.coordy(idx)]

        lineinfo = lineinfo[::-1, ...]
        out[N:2 * N, :2] = lineinfo[:, [0, 2]]
        for i in range(N):
            idx = lineinfo[i, 4]
            out[N + i, 2:] = [segy.coordx(idx), segy.coordy(idx)]

        out[-1] = out[0]
    else:
        raise ValueError("`mode` only support 'raw' and 'geom'")

    if not isinstance(fname, Pysegy):
        segy.close()

    # HACK: add more mode
    return out


def trace_count(fname: str) -> int:
    """
    Count the total numbers of a segy file

    Parameters
    ----------
    segy: str
        input segy file

    Returns
    -------
    int
        The total numbers of a segy file
    """
    if isinstance(fname, Pysegy):
        return fname.ntrace
    segy = Pysegy(str(fname))
    count = segy.ntrace
    segy.close()
    return count


def full_scan(fname: str,
              iline: int = None,
              xline: int = None,
              offset: int = 37,
              keys: np.ndarray = None,
              is4d: bool = None) -> dict:
    """
    Scan all keys of the SEG-Y file. This is useful for unsorted SEG-Y file.
    """
    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))

    if keys is None:
        if iline is None or xline is None:
            raise ValueError("keys is None, so iline and xline must be inputed, but got None")
        keys = segy.get_trace_keys([iline, xline, offset], [4] * 3, 0, segy.ntrace)
    else:
        assert keys.ndim == 2
        if keys.shape[1] == 2:
            is4d = False 
        elif keys.shape[1] == 3:
            is4d = True
        else:
            raise ValueError("keys' shape must be (N, 2) or (N, 3)")

    ib = keys[:, 0].min()
    ie = keys[:, 0].max()
    diff = np.diff(np.sort(keys[:, 0]))
    diff = diff[diff != 0]
    istep = diff.min()
    if (ie - ib) % istep != 0:
        raise RuntimeError("can not create geomtry (error when determine iline/istep)") # yapf: disable

    xb = keys[:, 1].min()
    xe = keys[:, 1].max()
    diff = np.diff(np.sort(keys[:, 1]))
    diff = diff[diff != 0]
    xstep = diff.min()
    if (xe - xb) % xstep != 0:
        raise RuntimeError("can not create geomtry (error when determine xline/xstep)") # yapf: disable

    if is4d is None or is4d == True:
        is4d = True
        ob = keys[:, 2].min()
        oe = keys[:, 2].max()
        if oe == ob:
            is4d = False
        elif np.unique(keys[:, 2]).size > 500:
            is4d = False
        else:
            try:
                diff = np.diff(np.sort(keys[:, 2]))
                diff = diff[diff != 0]
                ostep = diff.min()
                if (oe - ob) % ostep != 0:
                    raise RuntimeError("can not create geomtry (error when determine xline/xstep)") # yapf: disable

                no = int((oe - ob) // ostep + 1)
            except:
                is4d = False

    ni = int((ie - ib) // istep + 1)
    nx = int((xe - xb) // xstep + 1)
    nt = segy.nt
    location = [iline, xline, offset] if is4d else [iline, xline]
    shape = [ni, nx, no, nt] if is4d else [ni, nx, nt]
    ir = dict(min_iline=ib, max_iline=ie, istep=istep)
    xr = dict(min_xline=xb, max_xline=xe, xstep=xstep)

    keys[:, 0] = (keys[:, 0] - ib) / istep
    keys[:, 1] = (keys[:, 1] - xb) / xstep
    if is4d:
        keys[:, 2] = (keys[:, 2] - ob) / ostep
    keys = np.round(keys).astype(np.int32)

    geom = np.full(shape[:-1], -1, np.int32)
    if is4d:
        geom[keys[:, 0], keys[:, 1], keys[:, 2]] = np.arange(segy.ntrace)
    else:
        geom[keys[:, 0], keys[:, 1]] = np.arange(segy.ntrace)

    geominfo = {
        'location': location,
        'shape': shape,
        'iline': ir,
        'xline': xr,
    }
    if is4d:
        geominfo['offset'] = dict(min_offset=ob, max_offset=oe, ostep=ostep)

    geominfo['geom'] = geom

    return geominfo


def load_by_geom(
    fname,
    geominfo: dict,
    *,
    ib: int = 0,
    ie: int = -1,
    xb: int = 0,
    xe: int = -1,
    ob: int = 0,
    oe: int = -1,
    tb: int = 0,
    te: int = -1,
) -> np.ndarray:
    """
    Using the given geom, extract data without scanning. This is useful for unsorted SEG-Y file.
    
    `geominfo` must contain the key locations, shape, and ranges. For example:
    
    geominfo = {
        'location': [189, 193],
        'shape': [650, 781, 951],
        'iline': dict(min_iline=2201, max_iline=2850, istep=1),
        'xline': dict(min_xline=5650, max_xline=7210, xstep=2),
        'geom': geom, # geom is a (ni, nx) array for 3D or (ni, nx, no) for 4D}
    """
    is4d = False
    shape = geominfo['shape']
    if 'offset' in geominfo:
        is4d = True
        assert len(shape) == 4
    else:
        assert len(shape) == 3

    ie = shape[0] if ie == -1 else ie
    xe = shape[1] if xe == -1 else xe
    te = shape[3] if te == -1 else te
    if is4d:
        oe = shape[2] if oe == -1 else oe

    # check bound
    assert ib >= 0 and ib < ie and ie <= shape[0]
    assert xb >= 0 and xb < xe and xe <= shape[1]
    assert tb >= 0 and tb < te and te <= shape[3]
    if is4d:
        assert ob >= 0 and ob < oe and oe <= shape[2]

    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))

    if not is4d:
        x, y = np.meshgrid(np.arange(ib, ie), np.arange(xb, xe), indexing='ij')
        shape = x.shape
        index = geominfo['geom'][x.flatten(), y.flatten()]
        d = segy.collect(index, tb, te).reshape(*shape, -1)
    else:
        x, y, z = np.meshgrid(np.arange(ib, ie),
                              np.arange(xb, xe),
                              np.arange(ob, oe),
                              indexing='ij')
        shape = x.shape
        index = geominfo['geom'][x.flatten(), y.flatten(), z.flatten()]
        d = segy.collect(index, tb, te).reshape(*shape, -1)

    return d


################# atomic operation ####################

def set_keyi2(header: np.ndarray, loc: int, value: int, endian: str = '>'):
    loc = loc - 1
    header[loc:loc + 2] = np.array([value], dtype=f'{endian}i2').view(np.uint8)
    return header

def set_keyi4(header: np.ndarray, loc: int, value: int, endian: str = '>'):
    loc = loc - 1
    header[loc:loc + 4] = np.array([value], dtype=f'{endian}i4').view(np.uint8)
    return header

def get_keyi2(header: np.array, loc: int, endian: str = '>'):
    loc = loc - 1
    return np.frombuffer(header[loc:loc + 2], dtype=f'{endian}i2')[0]

def get_keyi4(header: np.array, loc: int, endian: str = '>'):
    loc = loc - 1
    return np.frombuffer(header[loc:loc + 4], dtype=f'{endian}i4')[0]



############### Deprecated functions ####################


def fromfile_by_guess(*args, **kwargs):
    raise RuntimeError(
        "`fromfile_by_guess` is deprecated and removed. Please use `fromfile` instead."
    )


def tofile_by_guess(*args, **kwargs):
    raise RuntimeError(
        "`tofile_by_guess` is deprecated and removed. Please use `tofile` instead."
    )


def scan_prestack(*args, **kwargs):
    raise RuntimeError(
        "`scan_prestack` is deprecated and removed. Please use `metaInfo` instead."
    )


def load_prestack3D(*args, **kwargs):
    raise RuntimeError(
        "`load_prestack3D` is deprecated and removed. Please use `fromfile` instead."
    )


def create_by_sharing_header_guess(*args, **kwargs):
    raise RuntimeError(
        "`create_by_sharing_header_guess` is deprecated and removed. Please use `create_by_sharing_header` instead."
    )


def scan_unsorted3D(*args, **kwargs):
    """
    Scan an unsored 3D SEG-Y file and get the geometry
    """
    raise RuntimeError(
        "`scan_unsorted3D` is deprecated and removed. Please use `full_scan` instead."
    )


def load_unsorted3D(*args, **kwargs):
    """
    using fromfile_without_scan to load unsorted 3D segy file
    """
    raise RuntimeError(
        "`load_unsorted3D` is deprecated and removed. Please use `load_by_geom` instead."
    )



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.
"""
Common SEG-Y file information, including header parsing details and encoding conversion dictionaries.

This module contains essential constants and information related to SEG-Y files, which are 
widely used in geophysical and seismic data storage. It provides details for parsing SEG-Y headers 
and includes conversion dictionaries for ASCII and EBCDIC encodings.
"""

# fmt: off
kBinaryHeaderHelp = {
    1: ("Job ID", 4),
    5: ("Line number", 4),
    9: ("Reel Number", 4),
    13: ("N traces per ensembel", 2),
    15: ("N auxiliary traces per ensembel", 2),
    17: ("Sample interval(dt)", 2),
    19: ("dt of original", 2),
    21: ("N samples per traces(ns)", 2),
    23: ("ns of orignal", 2),
    25: ("Data sample format code (1-IBM, 5-IEEE)", 2),
    27: ("Ensemble fold", 2),
    29: ("Trace sorting code", 2),
    31: ("vertical sum code", 2),
    33: ("Sweep freq at start(Hz)", 2),
    35: ("Sweep freq at end(HZ)", 2),
    37: ("Sweep length(ms)", 2),
    39: ("Sweep type code", 2),
    41: ("Trace number of sweep channel", 2),
    43: ("Sweep trace taper length in ms at strat", 2),
    45: ("Sweep trace taper length in ms at end", 2),
    47: ("Taper type", 2),
    49: ("Correlated data traces", 2),
    51: ("Binary gain recovered", 2),
    53: ("Amplitude recovery method", 2),
    55: ("Measurement system (units)", 2),
    57: ("Impulse signal polarity", 2),
    59: ("Vibratory polariy code", 2),
    61: ("Extended number of data traces per ensemble", 4),
    65: ("Extended number of auxiliary traces per ensemble", 4),
    69: ("Extended number of samples per data trace", 4),
    73: ("Extended sample interval, IEEE double precision (64-bit)", 9),
    81: ("Extended sample interval of original field recording, IEEE double precision (64-bit)", 9),
    89: ("Extended number of samples per data trace in original recording", 4),
    93: ("Extended ensemble fold", 4),
    97: ("The integer constant 16909060_10 (01020304_16)", 4),
    101: ("Unassigned", 200),
    301: ("Major SEG-Y Format Revision Number. This is an 8-bit unsigned value", 1),
    302: ("Minor SEG-Y Format Revision Number. This is an 8-bit unsigned value with a radix point between the first and second bytes.", 1),
    303: ("Fixed length trace flag", 2),
    305: ("Number of 3200-byte, Extended Textual File Header records following the Binary Header", 2),
    307: ("Max number of additional 240-byte trace header", 4),
    311: ("Time basis code", 2),
    313: ("number of trace header in this file, 64-bit unsigned integer value", 9),
    321: ("Byte offset of first trace relative to start of file or stream if known, otherwise zero. (64-bit unsigned integer value) ", 9),
    329: ("Number of 3200-byte data trailer stanza records following the last trace (4 byte signed integer).", 4),
    333: ("Unassigned", 68),
}

kTraceHeaderHelp = {
    1: ("Trace sequence number within line", 4),
    5: ("Trace sequence number within SEG-Y file", 4),
    9: ("Original field record number", 4),
    13: ("Trace number within the original field record", 4),
    17: ("Energy source point number", 4),
    21: ("Ensemble number", 4),
    25: ("Trace number within the ensemble", 4),
    29: ("Trace identification code", 2),
    31: ("Number of vertically summed traces yielding this trace", 2),
    33: ("Number of horizontally stacked traces yielding this trace", 2),
    35: ("Data use for", 2),
    37: ("Distance from center of the source point to the center of the receiver group", 4),
    41: ("Elevation of receiver group", 4),
    45: ("Surface elevation at source location", 4),
    49: ("Source depth below surface", 4),
    53: ("Seismic Datum elevation at receiver group", 4),
    57: ("Seismic Datum elevation at source", 4),
    61: ("Water column height at source location", 4),
    65: ("Water column height at receiver group location", 4),
    69: ("Scalar to be applied to all elevations and depths", 2),
    71: ("Scalar to be applied to all coordinates", 2),
    73: ("Source coordinate - X", 4),
    77: ("Source coordinate - Y", 4),
    81: ("Group coordinate - X", 4),
    85: ("Group coordinate - Y", 4),
    89: ("Coordinate units", 2),
    91: ("Weathering velocity", 2),
    93: ("Subweathering velocity", 2),
    95: ("Uphole time at source in ms", 2),
    97: ("Uphole time at group in ms", 2),
    99: ("Source static correction in ms", 2),
    101: ("Group static correction in ms", 2),
    103: ("Total static applied in ms", 2),
    105: ("Lag time A", 2),
    107: ("Lag Time B", 2),
    109: ("Delay recording time", 2),
    111: ("Mute time - Start time in ms", 2),
    113: ("Mute time - End time in ms", 2),
    115: ("Number of samples in this trace", 2),
    117: ("Sample interval for this trace", 2),
    119: ("Gain type of field instruments", 2),
    121: ("Instrument gain constant", 2),
    123: ("Instrument gain constant", 2),
    125: ("Correlated", 2),
    127: ("Sweep frequency at start", 2),
    129: ("Sweep frequency at end", 2),
    131: ("Sweep length in ms", 2),
    133: ("Sweep type", 2),
    135: ("Sweep trace taper length at start in ms", 2),
    137: ("Sweep trace taper length at end in ms", 2),
    139: ("Taper type", 2),
    141: ("Alias filter frequency (Hz)", 2),
    143: ("Alias filter slope (dB/octave)", 2),
    145: ("Notch filter frequency (Hz)", 2),
    147: ("Notch filter slope (dB/octave)", 2),
    149: ("Low-cut frequency (Hz)", 2),
    151: ("High-cut frequency (Hz)", 2),
    153: ("Low-cut slope (dB/octave)", 2),
    155: ("High-cut slope (dB/octave)", 2),
    157: ("Year data recorded", 2),
    159: ("Day of year", 2),
    161: ("Hour of day", 2),
    163: ("Minute of hour", 2),
    165: ("Second of minute", 2),
    167: ("Time basis code", 2),
    169: ("Trace weighting factor", 2),
    171: ("Geophone group number of roll switch position one", 2),
    173: ("Geophone group number of trace number one within original field record", 2),
    175: ("Geophone group number of last trace within original field record", 2),
    177: ("Gap size (total number of groups dropped)", 2),
    179: ("Over travel associated with taper at beginning or end of line", 2),
    181: ("X coordinate", 4),
    185: ("Y coordinate", 4),
    189: ("The in-line number", 4),
    193: ("The cross-line number", 4),
    197: ("Shotpoint number", 4),
    201: ("Scalar to be applied to the shotpoint number", 2),
    203: ("Trace value measurement unit", 2),
    205: ("Transduction Constant", 6),
    211: ("Transduction Units", 2),
    213: ("Device/Trace Identifier", 2),
    215: ("Scalar to be applied to bytes 95-114", 2),
    217: ("Source Type/Orientation", 2),
    219: ("Source Energy Direction with respect to the source orientation", 6),
    225: ("Source Measurement - Describes the source effort used to generate the trace", 6),
    231: ("Source Measurement Unit", 2),
    233: ("Either binary zeros or chars SEG00000", 8),
}

kTraceSortingHelp = {
    -1: "Other",
    0: "Unknown",
    1: "As recorded (no sorting)",
    2: "CDP ensemble",
    3: "Single fold continuous profile",
    4: "Horizontally stacked",
    5: "Common source point",
    6: "Common receiver point",
    7: "Common offset point",
    8: "Common mid-point",
    9: "Common conversion point",
}

kDataSampleFormatHelp = {
    1: "4-byte IBM floating-point",
    2: "4-byte, two's complement integer",
    3: "2-byte, two's complement integer",
    4: "4-byte fixed-point with gain (obsolete) (not implemented)",
    5: "4-byte IEEE floating-point",
    6: "8-byte IEEE floating-point (not implemented)",
    7: "3-byte two's complement integer (not implemented)",
    8: "1-byte, two's complement integer",
    9: "8-byte, two's complement integer (not implemented)",
    10: "4-byte, unsigned integer",
    11: "2-byte, unsigned integer",
    12: "8-byte, unsigned integer (not implemented)",
    15: "3-byte, unsigned integer (not implemented)" ,
    16: "1-byte, unsigned integer"
}

kEBCDICtoASCIImap = {
    64: ' ', 75: '.', 76: '<', 77: '(', 78: '+', 79: '|',
    80: '&', 90: '!', 91: '$', 92: '*', 93: ')', 94: ';',
    96: '-', 97: '/', 106: '|', 107: ',', 108: '%', 109: '_',
    110: '>', 111: '?', 121: '`', 122: ':', 123: '#', 124: '@',
    125: '\'', 126: '=', 127: '"', 129: 'a', 130: 'b', 131: 'c',
    132: 'd', 133: 'e', 134: 'f', 135: 'g', 136: 'h', 137: 'i',
    145: 'j', 146: 'k', 147: 'l', 148: 'm', 149: 'n', 150: 'o',
    151: 'p', 152: 'q', 153: 'r', 161: '~', 162: 's', 163: 't',
    164: 'u', 165: 'v', 166: 'w', 167: 'x', 168: 'y', 169: 'z',
    192: '{', 193: 'A', 194: 'B', 195: 'C', 196: 'D', 197: 'E',
    198: 'F', 199: 'G', 200: 'H', 201: 'I', 208: '}', 209: 'J',
    210: 'K', 211: 'L', 212: 'M', 213: 'N', 214: 'O', 215: 'P',
    216: 'Q', 217: 'R', 224: '\\', 226: 'S', 227: 'T', 228: 'U',
    229: 'V', 230: 'W', 231: 'X', 232: 'Y', 233: 'Z', 240: '0',
    241: '1', 242: '2', 243: '3', 244: '4', 245: '5', 246: '6',
    247: '7', 248: '8', 249: '9'
}

kASCIItoEBCDICmap = {
    ' ': 64, '.': 75, '<': 76, '(': 77, '+': 78, '|': 79,
    '&': 80, '!': 90, '$': 91, '*': 92, ')': 93, ';': 94,
    '-': 96, '/': 97, '|': 106, ',': 107, '%': 108, '_': 109,
    '>': 110, '?': 111, '`': 121, ':': 122, '#': 123, '@': 124,
    '\'': 125, '=': 126, '"': 127, 'a': 129, 'b': 130, 'c': 131,
    'd': 132, 'e': 133, 'f': 134, 'g': 135, 'h': 136, 'i': 137,
    'j': 145, 'k': 146, 'l': 147, 'm': 148, 'n': 149, 'o': 150,
    'p': 151, 'q': 152, 'r': 153, '~': 161, 's': 162, 't': 163,
    'u': 164, 'v': 165, 'w': 166, 'x': 167, 'y': 168, 'z': 169,
    '{': 192, 'A': 193, 'B': 194, 'C': 195, 'D': 196, 'E': 197,
    'F': 198, 'G': 199, 'H': 200, 'I': 201, '}': 208, 'J': 209,
    'K': 210, 'L': 211, 'M': 212, 'N': 213, 'O': 214, 'P': 215,
    'Q': 216, 'R': 217, '\\': 224, 'S': 226, 'T': 227, 'U': 228,
    'V': 229, 'W': 230, 'X': 231, 'Y': 232, 'Z': 233, '0': 240,
    '1': 241, '2': 242, '3': 243, '4': 244, '5': 245, '6': 246,
    '7': 247, '8': 248, '9': 249
}

# fmt: on



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.
"""
Coordinate transformations between inline-xline and geodetic (x, y) coordinates.

This module provides functions for converting between seismic survey coordinates (inline-xline) 
and geodetic coordinates (x, y). 
"""

import numpy as np


def get_transform_metrix(p1: np.ndarray, p2: np.ndarray) -> np.ndarray:
    """
    Calculate affine transform metrix `H` using LSTSQ, and ignore w.

    p2 = H \cdot p1

    Parameters
    -----------
    p1 : array-like
        points in coordinate 1
    p2 : array-like
        points in coordinate 2

    Returns
    --------
    H : array-like
        shape is (3, 3)
    """
    p1 = np.array(p1)
    p2 = np.array(p2)
    assert len(p1) == len(p2)
    N = len(p1)
    assert N >= 3

    A = np.zeros((2 * N, 6), float)
    b = p2.flatten()
    for i in range(N):
        A[2 * i] = [p1[i, 0], p1[i, 1], 1, 0, 0, 0]
        A[2 * i + 1] = [0, 0, 0, p1[i, 0], p1[i, 1], 1]
    H, residuals, rank, s = np.linalg.lstsq(A, b, rcond=None)
    H = np.concatenate([H, [0, 0, 1]], 0)

    return H.reshape(3, 3)


def apply_transform(p: np.ndarray,
                    transform_metrix: np.ndarray,
                    inv: bool = False) -> np.ndarray:
    """
    apply affine transform to translate the coordinate
    """
    p = np.array(p)
    if p.ndim == 1:
        p = p[np.newaxis, :]
    if inv:
        transform_metrix = np.linalg.inv(transform_metrix)

    p = np.concatenate([p, np.ones((len(p), 1))], 1).T

    out = np.dot(transform_metrix, p)

    x, y, w = out
    x /= w
    y /= w

    return np.c_[x, y]



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.

from typing import List, Tuple
import numpy as np
from cigsegy.cpp import _CXX_SEGY
from cigsegy.transform import get_transform_metrix, apply_transform
from cigsegy.interp import arbitray_line
from cigsegy import utils, plot, tools, createtool
import warnings



class ScanMixin:
    _segy: _CXX_SEGY.Pysegy

    def _eval_range(self):
        if self.ntrace < 6000:
            p0 = self._segy.collect(0, self.ntrace, 0, self.nt)
            self._min = mi
            self._max = ma
            return

        s, e = self.ntrace - 4000, 4000
        p0 = self._segy.collect(0, e, 0, self.nt)
        mi, ma = p0.min(), p0.max()

        p0 = self._segy.collect(s, self.ntrace, 0, self.nt)
        mi = min(mi, p0.min())
        ma = max(ma, p0.max())

        p0 = self._segy.collect(
            self.ntrace // 3,
            self.ntrace // 3 + 4000,
            0,
            self.nt,
        )
        mi = min(mi, p0.min())
        ma = max(ma, p0.max())

        self._min = mi
        self._max = ma

    def _scan(self, keylocs=None):
        if keylocs is None:
            keylocs = utils.guess(self._segy)
        elif isinstance(keylocs, dict):
            keylocs = utils.guess(self._segy, **keylocs)
        else:
            keylocs = utils.guess(self._segy, *keylocs)
        self._segy.setLocations(*keylocs[:3])
        self._segy.setSteps(*keylocs[3:6])
        self._segy.setXYLocations(*keylocs[6:8])
        ndim = 4 if keylocs[-1] is True else 3
        self._segy.set_segy_type(ndim)
        self._segy.scan()
        self._keylocs = self._segy.get_keylocs()
        self._metainfo = self._segy.get_metainfo()
        self._metainfo = {**self._keylocs, **self._metainfo}
        utils.post_process_meta(self._segy, self._metainfo, False)
        self._ndim = self._segy.ndim
        self._shape3 = self._segy.shape
        # self._lineinfo = self._segy.get_lineInfo() # TODO:

    def _scan_unsorted(self, keylocs=None, keys=None, ndim=None):
        if keys is not None:
            geom = tools.full_scan(self._segy, keys=keys)
        elif keylocs is None:
            raise ValueError(
                "When 'as_unsorted' is True, either 'keylocs' must be set, or 'keys' must be provided."
            )
        else:
            if ndim is None:
                is4d = None
            else:
                is4d = ndim == 4
            if isinstance(keylocs, List):
                geom = tools.full_scan(self._segy, *keylocs[:3], is4d=is4d)
            else:
                offset = keylocs.get('offset', 37)
                geom = tools.full_scan(self._segy, keylocs['iline'],
                                       keylocs['xline'], offset)
        self.update_geometry(geom['geom'])
        self._parser_unsorted_infos(geom)

    def _parser_unsorted_infos(self, geom):
        self._keylocs = self._segy.get_keylocs()
        self._metainfo = self._segy.get_metainfo()

        self._shape3 = tuple(geom['shape'])
        self._ndim = len(self._shape3)

        loc = geom['location']
        self._keylocs = dict(iline=loc[0], xline=loc[1])
        self._keylocs['istep'] = geom['iline']['istep']
        self._keylocs['xstep'] = geom['xline']['xstep']
        if len(loc) == 3:
            self._keylocs['offset'] = loc[2]
            self._keylocs['ostep'] = geom['offset']['ostep']

        self._metainfo['ni'] = self._shape3[0]
        self._metainfo['start_iline'] = geom['iline']['min_iline']
        self._metainfo['end_iline'] = geom['iline']['max_iline']
        self._metainfo['nx'] = self._shape3[1]
        self._metainfo['start_xline'] = geom['xline']['min_xline']
        self._metainfo['end_xline'] = geom['xline']['max_xline']
        if len(self._shape3) == 4:
            self._metainfo['no'] = self._shape3[2]
            self._metainfo['start_offset'] = geom['offset']['min_offset']
            self._metainfo['end_offset'] = geom['offset']['max_offset']
        self._metainfo['ndim'] = self._ndim


class GeometryMixin:
    _segy: _CXX_SEGY.Pysegy

    def map_to_index(self, index):
        if self._geometry is None:
            raise RuntimeError("geometry is not created, Call `update_geometry` first") # yapf: disable
        if self.ndim == 2:
            raise RuntimeError("ndim is 2, unsupport this function") # yapf: disable

        index = np.array(index)
        assert index.ndim == 2
        if self.ndim == 3:
            assert index.shape[1] == 2
            index = self._geometry[index[:, 0], index[:, 1]]
        elif self.ndim == 4:
            assert index.shape[1] == 3
            index = self._geometry[index[:, 0], index[:, 1], index[:, 2]]
        return index

    def update_geometry(self, geom=None):
        if geom is not None:
            self._geometry = geom
            return
        warnings.warn("This may be slow...")
        geom = tools.full_scan(self._segy, self._keylocs['iline'],
                               self._keylocs['xline'], self._keylocs['offset'])
        self._geometry = geom['geom']

    def update_trans_matrix(self, xyic=None):
        if self.ndim == 2:
            raise RuntimeError("ndim is 2, unsupport this function")
        if xyic is None:
            xyic = tools.get_lineInfo(self._segy, mode='geom')
            xyic = xyic[:, [2, 3, 0, 1]]
        self._trans_matrix = get_transform_metrix(xyic[:, 2:], xyic[:, :2])

    def xy_to_ix(self, xy, zero_origin=True):
        """
        Convert the x/y to inline/crossline

        Parameters
        ----------
        xy : np.ndarray
            The x/y array, shape is (n, 2)
        zero_origin : bool, optional
            Whether the x/y is zero-based, by default True,
            if is False, the x/y will be added by the min_x/min_y
        """
        xy = np.array(xy)
        shape = xy.shape
        if xy.ndim == 1:
            xy = xy.reshape(1, -1)
        if self._trans_matrix is None:
            self.update_trans_matrix()
        ic = apply_transform(xy, self._trans_matrix, inv=True)
        if zero_origin:
            ic[:, 0] -= self._metainfo['start_iline']
            ic[:, 1] -= self._metainfo['start_xline']
        return np.round(ic, 2).reshape(shape)

    def ix_to_xy(self, ix, zero_origin=True):
        """
        Convert the inline/crossline to x/y

        Parameters
        ----------
        ix : np.ndarray
            The inline/crossline array, shape is (n, 2)
        zero_origin : bool, optional
            Whether the inline/crossline is zero-based, by default True,
            if is False, the inline/crossline will be added by the min_inline/min_crossline
        """
        ix = np.array(ix)
        shape = ix.shape
        if ix.ndim == 1:
            ix = ix.reshape(1, -1)
        if zero_origin:
            ix[:, 0] += self._metainfo['start_iline']
            ix[:, 1] += self._metainfo['start_xline']
        if self._trans_matrix is None:
            self.update_trans_matrix()
        return np.round(apply_transform(ix, self._trans_matrix), 2).reshape(shape) # yapf: disable


class PlotMixin:
    _segy: _CXX_SEGY.Pysegy

    def plot_region(self, mode='line'):
        """
        plot the region map (x and y axis are inline and crossline)

        Parameters
        -----------
        mode : str
            one of ['line', 'cdpxy', 'xy'], default is 'line'
        """
        plot.plot_region(self._segy, mode)

    def plot_trace_keys(self,
                        keyloc: int,
                        beg: int = 0,
                        end: int = 1000) -> None:
        """
        plot the values (at keyloc in each trace) of the traces 
        range from beg to end .
        """
        plot.plot_trace_keys(self._segy, keyloc, beg, end)

    def plot_trace_keys2(
        self,
        beg: int = 0,
        end: int = 1000,
        k1: int = None,
        k2: int = None,
    ):
        """
        plot the values of the traces at k1 and k2, range from beg to end.
        If k1 and k2 is None, will plot iline/xline
        """
        if k1 is None:
            k1 = self._keylocs['iline']
        if k2 is None:
            k2 = self._keylocs['xline']
        plot.plot_trace_ix(self._segy, k1, k2, beg, end)

    def plot_trace_keys3(
        self,
        beg: int = 0,
        end: int = 1000,
        k1: int = None,
        k2: int = None,
        k3: int = None,
    ):
        """
        plot the values of the traces at k1, k2 and k3, range from beg to end.
        If k1, k2 and k3 is None, will plot iline/xline/offset
        """
        if k1 is None:
            k1 = self._keylocs['iline']
        if k2 is None:
            k2 = self._keylocs['xline']
        if k3 is None:
            k3 = self._keylocs['offset']
        plot.plot_trace_ixo(self._segy, k1, k2, k3, beg, end)

    def plot3d(self, use_viser=False):
        """
        plot 3d
        """
        assert self.ndim == 3, "The data is not 3D"
        if not use_viser:
            try:
                import cigvis
            except ImportError:
                raise ImportError("To use this function, you need to install cigvis")
            nodes = cigvis.create_slices(self)
            cigvis.plot3D(nodes)
        else:
            try:
                from cigvis import viserplot
            except ImportError:
                raise ImportError("To use this function, you need to install cigvis['viser']")
            nodes = viserplot.create_slices(self)
            viserplot.plot3D(nodes)


class InterpMixin:
    _segy: _CXX_SEGY.Pysegy

    def arbitrary_line(self, points, ptype='auto', return_path=True, di=1):
        """
        Extract an arbitrary line from the the SEG-Y file. 
        The path is consisted by points

        Parameters
        -----------
        points : ArrayLike
            shape is (N, 2), it also can be a list
        ptype : str
            one of ['auto', 'zero', 'line', 'xy'], default is 'auto'.
            'zero' means points are taken from a geometry with zero-origin (i.e., numpy array indexes), 
            'line' means points are taken from the inline/xline geometry, 
            'xy' means points are taken from the X-Y geometry.
        return_path : bool
            if true, will return the path
        di : float
            the interval between two points of the path
        """
        points = self._process_points(points, ptype)
        out, p, indices = arbitray_line(self, points, di)
        if not return_path:
            return out
        return out, p, indices

    def extract_arbitrary_line_by_view(self,
                                       bmap: str = 'data',
                                       draw_arb: bool = True,
                                       *,
                                       return_values: bool = True,
                                       line: bool = True,
                                       idx: int = 50,
                                       cline='#F3AA3C'):
        """
        extract arbitrary line from seismic data by clicking

        Parameters
        ----------
        - bmap: str
            background map, 'data' or 'blank'
        - line : bool
            whether to draw the broken line 
        - idx: int 
            the slice index of the seismic data if bmap is 'data'
        - cline: str
            color of the line

        Returns
        -------
        - out: np.ndarray
            extracted arbitrary line
        - p: np.ndarray
            extracted arbitrary line path
        - coords: np.ndarray
            the coordinates by clicking
        """
        out = plot.extract_arbitrary_line_by_view(
            self,
            bmap,
            draw_arb,
            line=line,
            idx=idx,
            cline=cline,
        )
        if return_values:
            return out

    def align_coordinates(self, fname: str):
        """
        Interpolate the input SEG-Y data to align with the self coordinate system.
        """
        raise NotImplementedError("Not Implemented yet")

    def _process_points(self, points, ptype='auto'):
        assert self.ndim == 3, "The data is not 3D"
        points = np.array(points)
        # HACK: Need optimize
        if ptype == 'auto':
            if points.max() > 100000:
                ptype = 'xy'
            elif points[:, 0].max() >= self.shape[0] or points[:, 1].max() >= self.shape[1]: # yapf: disable
                ptype = 'line'
            else:
                ptype = 'zero'

        if ptype == 'line':
            points[:, 0] -= self.iline_range[0]
            points[:, 1] -= self.xline_range[0]
        elif ptype == 'xy':
            points = self.xy_to_ix(points)

        self._check_bound(0, points[:, 0].min(), points[:, 0].max())
        self._check_bound(1, points[:, 1].min(), points[:, 1].max())
        return points


class RWMixin:
    _segy: _CXX_SEGY.Pysegy

    def to_numpy(self):
        """like pandas"""
        self._segy.show_progress(True)
        if self.unsorted and self.ndim > 2:
            d = self[...]
        else:
            d = self._segy.read()
        self._segy.show_progress(False)
        return d

    def tofile(self, fpath: str, load: bool = True):
        """
        save the SEG-Y file to a binary file without headers

        Parameters
        -----------
        fpath : str
            the save path
        load : bool
            if load is true, will load the file into memery first, then write to a file
        """
        self._segy.show_progress(True)
        if load:
            self.to_numpy().tofile(fpath)
        else:
            self._segy.tofile(fpath, self.ndim == 2)
        self._segy.show_progress(False)

    # fmt: off
    def _read4d(self, idx) -> np.ndarray:
        assert self.ndim == 4, "The data is not 4D"
        self._check_bound2(idx)

        num = idx[1::2].count(-1)

        if num == 0 and not self.unsorted:
            d = self._segy.read4d(*idx)
        else:
            if not self.is_create_geometry:
                raise RuntimeError("Need create the geometry first, please call `update_geometry` first")
            grid, shape = self._create_meshgrid(idx[:-2])
            tidx = self.map_to_index(grid)
            if idx[-1] is None:
                d = self._segy.collect(tidx, 0, self.nt).reshape(*shape, -1)
                d = d[..., idx[-2]]
            else:
                d = self._segy.collect(tidx, idx[-2], idx[-1]).reshape(*shape, -1)

        return self._post_process(d)


    def _read3d(self, idx) -> np.ndarray:
        assert self.ndim == 3, "The data is not 3D"
        self._check_bound2(idx)
        num = idx[1::2].count(-1)
        if num == 0 and not self.unsorted:
            if self._fast_read and self._is_time_slice(idx):
                d = self._segy.read_tslice(idx[4], self._fstep[0], self._fstep[1])
            else:
                d = self._segy.read3d(*idx)
        else:
            if not self.is_create_geometry:
                raise RuntimeError("Need create the geometry first, please call `update_geometry` first")
            grid, shape = self._create_meshgrid(idx[:-2])
            tidx = self.map_to_index(grid)
            if idx[-1] is None:
                d = self._segy.collect(tidx, 0, self.nt).reshape(*shape, -1)
                d = d[..., idx[-2]]
            else:
                d = self._segy.collect(tidx, idx[-2], idx[-1]).reshape(*shape, -1)

        return self._post_process(d)


    def _read2d(self, idx) -> np.ndarray:
        assert self.ndim == 2, "The data is not 2D"
        self._check_bound2(idx)

        # the first dim is ndarray
        if idx[1] is None:
            # the time dim is ndarray
            if idx[-1] is None:
                data = self._segy.collect(idx[0], 0, self.nt)
                data = data[:, idx[2]]
            else:
                data = self._segy.collect(idx[0], idx[2], idx[3])
        else:  # the first dim is ib, ie
            if idx[-1] is None:
                data = self._segy.collect(idx[0], idx[1], 0, self.nt)
                data = data[:, idx[2]]
            else:
                data = self._segy.collect(idx[0], idx[1], idx[2], idx[3])

        return self._post_process(data)


    def _write4d(self, idx, data) -> None:
        assert self.ndim == 4, "The data is not 4D"
        self._check_bound2(idx)
        self._check_wmode_lastdim(idx)
        self._check_data_shape(data, idx)

        num = idx[1::2].count(-1)

        if num == 0 and not self.unsorted:
            self._segy.write4d(data, *idx)
        else:
            if not self.is_create_geometry:
                raise RuntimeError("Need create the geometry first, please call `update_geometry` first")
            grid, shape = self._create_meshgrid(idx[:-2])
            tidx = self.map_to_index(grid)
            self._segy.write_traces(data, tidx, idx[-2], idx[-1])


    def _write3d(self, idx, data) -> None:
        assert self.ndim == 3, "The data is not 3D"
        self._check_bound2(idx)
        self._check_wmode_lastdim(idx)
        self._check_data_shape(data, idx)

        num = idx[1::2].count(-1)

        if num == 0 and not self.unsorted:
            self._segy.write3d(data, *idx)
        else:
            if not self.is_create_geometry:
                raise RuntimeError("Need create the geometry first, please call `update_geometry` first")
            grid, shape = self._create_meshgrid(idx[:-2])
            tidx = self.map_to_index(grid)
            self._segy.write_traces(data, tidx, idx[-2], idx[-1])


    def _write2d(self, idx, data) -> None:
        assert self.ndim == 2, "The data is not 2D"
        assert data.ndim <= 2, "The data is not 2D"
        self._check_bound2(idx)
        self._check_wmode_lastdim(idx)
        self._check_data_shape(data, idx)

        # the first dim is ndarray
        if idx[1] is None:
            # the time dim is ndarray
            self._segy.write_traces(data, idx[0], idx[2], idx[3])
        else:  # the first dim is ib, ie

            self._segy.write_traces(idx[0], idx[1], idx[2], idx[3])


    def _create_meshgrid(self, idx):
        if idx[1] is None:
            x = idx[0]
        else:
            x = np.arange(idx[0], idx[1])
        if idx[3] is None:
            y = idx[2]
        else:
            y = np.arange(idx[2], idx[3])

        if self.ndim == 3:
            X, Y = np.meshgrid(x, y, indexing='ij')
            shape = X.shape
            return np.c_[X.flatten(), Y.flatten()], shape
        elif self.ndim == 4:
            if idx[5] is None:
                z = idx[4]
            else:
                z = np.arange(idx[4], idx[5])
            X, Y, Z = np.meshgrid(x, y, z, indexing='ij')
            shape = X.shape
            return np.c_[X.flatten(), Y.flatten(), Z.flatten()], shape
        else:
            raise RuntimeError("only support ndim == 3 or ndim == 4")


    def _post_process(self, d: np.ndarray):
        d = d.squeeze()
        if d.ndim == 1 and d.size == 1:
            return d[0]
        elif d.ndim == 0:
            return float(d)
        return d


    def _process_keys(self, key) -> List:
        if isinstance(key, (int, np.integer)):

            if key < 0:
                key += self.shape[0]
            if key < 0 or key >= self.shape[0]:
                raise IndexError("Index out of range")

            if self.ndim == 2:
                return key, key + 1, 0, self.shape[1]
            elif self.ndim == 3:
                return key, key + 1, 0, self.shape[1], 0, self.shape[2]
            else:
                return key, key + 1, 0, self.shape[1], 0, self.shape[2], 0, self.shape[3]

        elif key is Ellipsis:

            if self.ndim == 2:
                return 0, self.shape[0], 0, self.shape[1]
            elif self.ndim == 3:
                return 0, self.shape[0], 0, self.shape[1], 0, self.shape[2]
            else:
                return 0, self.shape[0], 0, self.shape[1], 0, self.shape[2], 0, self.shape[3]

        elif isinstance(key, Tuple):

            return self._process_keys_tuple(key)

        elif isinstance(key, (slice, List, np.ndarray)):

            if isinstance(key, slice):
                ib = 0 if key.start is None else key.start
                ie = self.shape[0] if key.stop is None else key.stop
                assert key.step is None, "step is not supported"
            else:
                ib, ie = np.array(key), None

            if self.ndim == 2:
                return ib, ie, 0, self.shape[1]
            elif self.ndim == 3:
                return ib, ie, 0, self.shape[1], 0, self.shape[2]
            else:
                return ib, ie, 0, self.shape[1], 0, self.shape[2], 0, self.shape[3]

        else:
            raise IndexError("Invalid index slices")


    def _process_keys_tuple(self, key) -> List:
        if len(key) > self.ndim:
            raise IndexError(f"Too many dimensions: expected at most {self.ndim}, got {len(key)}")

        num_ellipsis = key.count(Ellipsis)
        if num_ellipsis > 1:
            raise ValueError("Only one ellipsis (...) allowed")

        if num_ellipsis == 1:
            idx = key.index(Ellipsis)
            n_insert = self.ndim - len(key) + 1
            key = key[:idx] + (slice(None),) * n_insert + key[idx + 1:]

        start_idx = [None] * self.ndim
        end_idx = [None] * self.ndim
        for i, k in enumerate(key):
            if k is None:
                continue

            if isinstance(k, (int, np.integer)):
                if k < 0:
                    k += self.shape[i]
                start_idx[i] = k
                end_idx[i] = k + 1

            elif isinstance(k, slice):
                if not (k.step is None or k.step == 1):
                    raise IndexError(f"only support step is 1, while got a step {k.step} in the {i}th dimension")

                start_idx[i] = k.start or 0
                end_idx[i] = k.stop or self.shape[i]

            elif isinstance(k, (List, np.ndarray)):
                start_idx[i] = np.array(k)
                end_idx[i] = -1

            else:
                raise IndexError("Invalid index slices")

        out = []
        for i in range(self.ndim):
            if start_idx[i] is None:
                out.append(0)
            else:
                out.append(start_idx[i])

            if end_idx[i] is None:
                out.append(self.shape[i])
            else:
                out.append(end_idx[i])

        return out

    def _is_time_slice(self, idx):
        # Only used in _read3d and num==0 and not unsorted
        if idx[1] - idx[0] == self.shape[0] and idx[3] - idx[2] == self.shape[1] and idx[5]-idx[4]==1: # yapf: disable
            return True
        return False


class CheckMixin:

    def _check_bound(self, dim, ib, ie):
        if ie == None:
            assert isinstance(ib, np.ndarray) and ib.ndim == 1, f"if ie is 0, ib must be a 1D numpy array"
            assert ib.min() >= 0 and ib.max() <= self.shape[dim], f"index array out of range in dim {dim}"
        else:
            assert ib >= 0 and ib < ie and ib <= self.shape[dim], f"index out of range in dim {dim}"


    def _check_bound2(self, idx):
        assert len(idx) == self.ndim * 2, f"ndim is {self.ndim}, need {self.ndim*2} idx, but got {len(idx)}"
        for i in range(self.ndim):
            self._check_bound(i, idx[i*2], idx[i*2+1])


    def _check_bound_idx(self, index, dim):
        assert isinstance(index, (int, np.integer)), "index must be int"
        assert index >= 0 and index < self.shape[dim], f"In dimension {dim}, index {index} out of range"

    def _check_wmode_lastdim(self, idx):
        if idx[-1] is None:
            raise TypeError("Indexing with a list or ndarray for the last dimension is not supported when writing.")

    def _check_data_shape(self, data: np.ndarray, idx):
        dstshape = []
        for i in range(self.ndim):
            if idx[i * 2 + 1] is None:
                dstshape.append(idx[i * 2].shape)
            else:
                dstshape.append(idx[i * 2 + 1] - idx[i * 2])
        dstshape = [k for k in dstshape if k != 1]
        assert tuple(dstshape) == data.squeeze().shape, "shape of the input data is not match the shape of the data to write"


class SegyCMixin:
    _segy: _CXX_SEGY.Pysegy

    def cut(self,
            outname: str,
            ranges: List,
            as2d: bool = False,
            textual: str = '') -> None:
        """
        cut the SEG-Y file into a new SEG-Y file
        """
        if as2d:
            assert len(ranges) == 4, "The length of `ranges` must be 4, as you set the `as2d` to True"
        else:
            assert len(ranges) == self.ndim * 2, f"ndim is {self.ndim}, need {self.ndim*2} idx, but got {len(ranges)}"

        textual = createtool.generate_textual(self._metainfo, textual)

        self._segy.cut(outname, ranges, as2d, textual)

    def create_by_sharing_header(self,
                                 outname: str,
                                 src,
                                 shape=None,
                                 *,
                                 start=None,
                                 as2d=False,
                                 textual='') -> None:
        if isinstance(src, np.ndarray):
            if shape is not None:
                warnings.warn("The shape is ignored, as the src is ndarray")
            shape = src.shape
        elif shape is None:
            raise ValueError("src (filename) is not ndarray, shape is None, need to specify the shape")

        if as2d:
            assert len(shape) == 2, "The shape must be 2D, as you set the `as2d` to True"

        if start is None:
            start = [0] * len(shape)

        if len(textual) > 0 and len(textual) != 3200:
            textual = createtool.generate_textual(textual)

        self._segy.show_progress(True)
        if isinstance(src, np.ndarray):
            self._segy.create_by_sharing_header(outname, src, start, as2d, textual)
        else:
            self._segy.create_by_sharing_header(outname, src, shape, start, as2d, textual)
        self._segy.show_progress(False)
        


class AccessMixin:
    _segy: _CXX_SEGY.Pysegy

    def __getattr__(self, name):
        if name in ['iline', 'xline', 'offset', 'coordx', 'coordy', 'itrace']:
            return self._SegyAccessor(self._segy, name)

        return super().__getattr__(name)


    def __setattr__(self, name: str, value) -> None:
        if name in ['iline', 'xline', 'offset', 'coordx', 'coordy', 'itrace']:
            accessor = self._SegyAccessor(self._segy, name)
            accessor[:] = value
            return
        else:
            super().__setattr__(name, value)

    class _SegyAccessor:

        def __init__(self, segy, attribute, mode='r'):
            assert mode in ['r', 'rw']
            self._segy = segy
            self.attribute = attribute
            self._mode = mode

        def __getitem__(self, index):
            index = self._process_index(index)
            method = getattr(self._segy, self.attribute)
            if isinstance(index, (int, np.integer)):
                return method(index)
            else:
                return np.array([method(i) for i in index])

        def __setitem__(self, index, value):
            if self._mode == 'r':
                raise RuntimeError("The SEG-Y file is not writable, as you set the `mode` to 'r'. If you want to enable write mode, set to `rw`") # yapf: disable
            index = self._process_index(index)
            # TODO: check the value size?

            if self.attribute == 'itrace':
                method = getattr(self._segy, f"write_{self.attribute}")
            else:
                method = getattr(self._segy, f"set_{self.attribute}")
            if isinstance(index, (int, np.integer)):
                method(index, value)
            else:
                for i, v in zip(index, value):
                    method(i, v)

        def _process_index(self, keys):
            if isinstance(keys, tuple):
                # Raise an error if more than one dimension is indexed
                raise IndexError("Only 1D indexing is supported.")

            if isinstance(keys, (int, np.integer)):
                if keys < 0:
                    keys += self._segy.ntrace
                return keys

            elif isinstance(keys, slice):
                # Slice index
                start, stop, step = keys.start, keys.stop, keys.step
                if keys.start < 0:
                    start += self._segy.ntrace
                if keys.stop < 0:
                    stop += self._segy.ntrace
                keys = np.arange(start, stop, step)
                assert keys.min() >= 0 and keys.max() < self._segy.ntrace, "Index out of range"
                return keys

            elif isinstance(keys, list) or isinstance(keys, np.ndarray):
                # List or numpy array of indices
                keys = np.array(keys).squeeze()
                assert keys.ndim == 1, "Only 1D indexing is supported."
                assert keys.min() >= 0 and keys.max() < self._segy.ntrace, "Index out of range"
                return keys

            else:
                raise TypeError("Unsupported index type.")


        def __repr__(self) -> str:
            d = np.array([self.__getitem__(i) for i in [0, 1, -2, -1]])
            if self.attribute == 'itrace':
                usage = "traces: \n"
                for i in range(4):
                    idx = i
                    if idx > 1:
                        idx = self._segy.ntrace - 4 + i
                    usage += f"trace {idx}: [{d[i, 0]:.4f}, {d[i, 1]:.4f}, {d[i, 2]:.4f}, {d[i, 3]:.4f}, ...]\n"
                    if i == 1:
                        usage += "...\n"
            else:
                usage = f"{self.attribute}: [{d[0]}, {d[1]}, ..., {d[2]}, {d[3]}]"
            return usage

    def bkeyi2(self, loc):
        """get binary header value at loc, view it as int16_t"""
        return self._segy.bkeyi2(loc)

    def bkeyi4(self, loc):
        """get binary header value at loc, view it as int32_t"""
        return self._segy.bkeyi4(loc)

    def keyi2(self, idx, loc):
        """get idx-th trace header value at loc, view it as int16_t"""
        return self._segy.keyi2(idx, loc)

    def keyi4(self, idx, loc):
        """get idx-th trace header value at loc, view it as int16_t"""
        return self._segy.keyi4(idx, loc)

    def set_bkeyi2(self, loc, value):
        """set binary header value at loc, view it as int16_t"""
        return self._segy.set_bkeyi2(loc, value)

    def set_bkeyi4(self, loc, value):
        """set binary header value at loc, view it as int32_t"""
        return self._segy.set_bkeyi4(loc, value)

    def set_keyi2(self, idx, loc, value):
        """set idx-th trace header value at loc, view it as int16_t"""
        return self._segy.set_keyi2(idx, loc, value)

    def set_keyi4(self, idx, loc, value):
        """set idx-th trace header value at loc, view it as int16_t"""
        return self._segy.set_keyi4(idx, loc, value)

    def textual_header(self, code='u', printtext=True):
        """get textual header"""
        t = self._segy.textual_header(code)
        if printtext:
            print(t)
        else:
            return t
# fmt: on


class InnerMixin:
    _segy: _CXX_SEGY.Pysegy

    def __array__(self):
        """To support np.array(SegyNP(xxx))"""
        return self.to_numpy()

    def __getitem__(self, slices) -> np.ndarray:
        idx = self._process_keys(slices)
        if self._ndim == 4:
            return self._read4d(idx)
        elif self._ndim == 3:
            return self._read3d(idx)
        else:
            return self._read2d(idx)

    def __setitem__(self, slices, data: np.ndarray) -> None:
        if data.dtype != np.float32:
            raise TypeError("The data type of the input data must be np.float32") # yapf: disable
        idx = self._process_keys(slices)
        if self._ndim == 4:
            return self._write4d(idx, data)
        elif self._ndim == 3:
            return self._write3d(idx, data)
        else:
            return self._write2d(idx, data)

    def __array_function__(self, func, types, args, kwargs):
        if func is np.min:
            return self.min()
        elif func is np.max:
            return self.max()
        elif func is np.nanmin:
            return self.min()
        elif func is np.nanmax:
            return self.max()
        elif func is np.save:
            fpath, obj = args
            np.save(fpath, self.__array__())
            return
        raise NotImplementedError(
            f"Function {func} is not implemented for SegyNP")

    def __del__(self):
        if hasattr(self, '_segy'):
            self._segy.close()

    def __repr__(self) -> str:
        out = f"cigsegy.SegyNP class, file name: '{self.file_name}'\n\n"
        keys = self._segy.get_keylocs()
        meta = self._segy.get_metainfo()
        meta = {**keys, **meta}
        meta = utils.post_process_meta(self._segy, meta)
        out += utils.parse_metainfo(meta)
        return out

    def __len__(self) -> int:
        return self.shape[0]


class SegyNP(InnerMixin, RWMixin, InterpMixin, PlotMixin, GeometryMixin,
             ScanMixin, CheckMixin, AccessMixin, SegyCMixin):

    def __init__(self,
                 filename: str,
                 keylocs: dict = None,
                 mode: str = 'r',
                 *,
                 ndim: int = None,
                 as_unsorted: bool = False,
                 fast_read: bool = False,
                 keys: dict = None) -> None:
        np.set_printoptions(suppress=True)

        assert mode in ['r', 'rw'], "`mode` only can be 'r' or 'rw'"
        self._ndim = ndim
        self._fname = filename

        self._segy = _CXX_SEGY.Pysegy(str(filename), mode=='rw')
        self._segy.show_progress(False)
        self._mode = mode
        if self._mode == 'rw':
            warnings.warn(
                "\n**Dangerous!!!** You are using a writable mode ('rw'), which may **alter** the SEG-Y file. \n"
                "It is strongly recommended to make a **backup copy** of the file before proceeding "
                "to avoid any potential irreversible changes.", UserWarning)

        self._fast_read = fast_read  # TODO: set as 'auto'? and how to set step?

        # for values
        self._min = None
        self._max = None
        self._eval_range()

        # for coordinates transform
        self._trans_matrix = None
        self._geometry = None
        self._north = None

        self._keylocs = None
        self._metainfo = None
        self._lineinfo = None
        self._fstep = (2, 2)

        self._shape2 = (self._segy.ntrace, self._segy.nt)
        self._shape3 = None

        if ndim == 2 and as_unsorted:
            warnings.warn("`ndim` is 2, so `as_unsorted` will be ignored")
            as_unsorted = False
        self._unsorted = as_unsorted

        if ndim is None or ndim != 2:
            if self._unsorted:
                self._scan_unsorted(keylocs, keys, ndim)
            else:
                try:
                    self._scan(keylocs)
                except Exception as e:
                    raise RuntimeError(f"{str(e)}\n This SEG-Y file may be unsorted, you can pass `as_unsorted` to view it as unsorted file, but it may be slow") from e # yapf: disable
            if ndim is not None and self.ndim != ndim:
                raise RuntimeError(f"You set ndim as {ndim}, but the SEG-Y file's ndim is {self.ndim}") # yapf: disable

    @property
    def ntrace(self) -> int:
        """
        Number of traces in the SEG-Y file
        """
        return self._segy.ntrace

    @property
    def ndim(self) -> int:
        """
        Number of dimensions of the data
        """
        return self._ndim

    @property
    def shape(self) -> Tuple:
        """
        the shape of the data
        """
        if self.ndim == 2:
            return tuple(self._shape2)
        else:
            return tuple(self._shape3)

    @property
    def dtype(self) -> np.dtype:
        """
        the data type of the data, always be np.float32
        """
        return np.float32

    @property
    def nt(self) -> int:
        """
        length of the time axis, i.e., the number of samples for each trace
        """
        return self._segy.nt

    @property
    def file_name(self) -> str:
        """
        the file name of the SEG-Y file
        """
        return self._fname

    @property
    def north(self):
        """
        the north direction of the SEG-Y file, only available for 3D/4D data
        """
        if self._north is None:
            N = self.ntrace // 3
            xy1 = [self.coordx[N], self.coordy[N]]
            xy2 = xy1.copy()
            xy2[1] += 1000
            n = self.xy_to_ix(np.array([xy1, xy2]))
            di = n[1] - n[0]
            self._north = di / np.linalg.norm(di)

        return self._north

    @property
    def unsorted(self) -> bool:
        """
        whether the SEG-Y file is unsorted
        """
        return self._unsorted

    @property
    def is_create_geometry(self) -> bool:
        """
        whether the geometry is created
        """
        return self._geometry is not None

    def close(self) -> None:
        """
        close the SEG-Y file
        """
        if hasattr(self, '_segy'):
            self._segy.close()

    def to_2d(self):
        """
        Treat the SEG-Y file as a collection of traces, shape is like (ntrace, nt)
        """
        self._ndim = 2

    def to_nd(self):
        """
        Treat the SEG-Y file as a 3D/4D array. If the SEG-Y file is scanned, the keylocs will be ignored 
        """
        if self._shape3 is None:
            self._scan()
        self._ndim = self._segy.ndim

    def max(self, real=False) -> float:
        """
        return the maximum value of the data, 
        if real is False, the maximum value is not real, and we just read a part of traces to calculate max
        if real is True, we read all traces to calculate max
        """
        if real:
            return self[...].min()
        return self._max

    def min(self, real=False) -> float:
        """
        return the min value of the data, 
        if real is False, the min value is not real, and we just read a part of traces to calculate min
        if real is True, we read all traces to calculate min
        """
        if real:
            return self[...].max()
        return self._min

    @property
    def lineinfo(self):
        return self._lineinfo

    @property
    def metainfo(self):
        return self._metainfo

    @property
    def keylocs(self):
        return self._keylocs

    @property
    def access_mode(self) -> str:
        return self._mode

    @property
    def fast_read(self) -> bool:
        return self._fast_read

    @fast_read.setter
    def fast_read(self, value: bool) -> None:
        self._fast_read = value

    def set_fast_read_steps(self, istep, xstep):
        self._fstep = (istep, xstep)



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.
"""
Interpolation functions for extracting arbitrary lines, resampling data, and data fusion.

This module provides a set of interpolation functions that are essential for handling seismic data 
and other geophysical datasets. These functions enable the extraction of arbitrary lines from data grids, 
resampling of data to different resolutions, and the fusion of multiple datasets.
"""

import numpy as np
from .transform import apply_transform
from cigsegy import ExceptionWrapper

try:
    from numba import njit
except ImportError:

    def njit(func):

        def wrapper(*args, **kwargs):
            print("Warning: Numba is not installed. This function may run slowly without Numba's optimization. To significantly improve performance, please install Numba by running `pip install numba`.") # yapf: disable
            return func(*args, **kwargs)

        return wrapper


try:
    from scipy.spatial import cKDTree
except BaseException as E:
    cKDTree = ExceptionWrapper(E, "run `pip install scipy` to install the the dependency, becuase we need use cKDTree") # yapf: disable


def merge_data(d1, d2, start2, useA=True):
    """
    Merge two data with different start points.
    coordinates range of d1 is (0, n1), (0, n2)
    coordinates range of d2 is (start2[0], start2[0]+n1), (start2[1], start2[1]+n2)
    """
    assert d1.ndim == 3 and d2.ndim == 3
    assert d1.shape[2] == d2.shape[2]

    mask1 = ~np.all(d1 == 0, axis=2)
    mask2 = ~np.all(d2 == 0, axis=2)
    n11, n21, n3 = d1.shape
    n12, n22, _ = d2.shape
    x0, y0 = start2

    x_min = min(0, x0)
    x_max = max(n11, x0 + n12)
    y_min = min(0, y0)
    y_max = max(n21, y0 + n22)

    nx_total = x_max - x_min
    ny_total = y_max - y_min

    x1s, y1s = -x_min, -y_min
    x2s, y2s = x0 - x_min, y0 - y_min

    mmask1 = np.zeros((nx_total, ny_total), dtype=bool)
    mmask2 = np.zeros((nx_total, ny_total), dtype=bool)
    mmask1[x1s:x1s + n11, y1s:y1s + n21] = mask1
    mmask2[x2s:x2s + n12, y2s:y2s + n22] = mask2

    if useA:
        merged_mask = np.logical_not(mmask1) & mmask2
    else:
        merged_mask = mmask2

    out = np.zeros((nx_total, ny_total, n3), dtype=d1.dtype)
    out[x1s:x1s + n11, y1s:y1s + n21] = d1

    # @njit
    # def _merge(out, d2, merged_mask, n12, n22, x2s, y2s):
    #     for i in range(n12):
    #         for j in range(n22):
    #             ii = i + x2s
    #             jj = j + y2s
    #             if merged_mask[ii, jj]:
    #                 out[ii, jj] = d2[i, j]
    #     return out
    # out = _merge(out, d2, merged_mask, n12, n22, x2s, y2s)

    I_grid, J_grid = np.meshgrid(np.arange(n12), np.arange(n22), indexing='ij')
    II = I_grid + x2s
    JJ = J_grid + y2s

    valid_mask = (II >= 0) & (II < out.shape[0]) & (JJ >= 0) & (JJ
                                                                < out.shape[1])

    combined_mask = merged_mask[II, JJ] & valid_mask

    I_selected = I_grid[combined_mask]
    J_selected = J_grid[combined_mask]
    II_selected = II[combined_mask]
    JJ_selected = JJ[combined_mask]

    out[II_selected, JJ_selected, :] = d2[I_selected, J_selected, :]

    return out


def align_coordinates(
    tgt: np.ndarray,
    transr: np.ndarray,
    transt: np.ndarray,
    xyicr: np.ndarray,
    xyict: np.ndarray,
    order: int = 1,
):
    """
    0. information of ref and tgt
        ref: xyr, icr_r, datar
        tgt: xyt, ict_t, datat
        icr_r means the ic coordinates of ref in ref geometry
        ict_r means the ic coordinates of tgt in ref geometry
    
    1. convert xyt to ict_r 
    2. generate gridt_r based on ict_r
    3. convert gridt_r to xy_gt
    4. convert xy_gt to ic_gt_t in tgt geometry
    5. interpolate data_gt_t using datat
    6. reshape data_gt_t to meet the 3d shape

    Parameters
    ----------
    tgt : SegyNP or np.ndarray
        target data
    transr : np.ndarray
        transform matrix from ic to xy of ref
    transt : np.ndarray
        transform matrix from ic to xy of tgt
    xyicr : np.ndarray
        shape is (N, 4), each row is [x, y, iline, xline] in ref geometry
    xyict : np.ndarray
        shape is (N, 4), each row is [x, y, iline, xline] in tgt geometry

    Returns
    -------
    np.ndarray
        shape is (N, n3)
    offset : List
        [x0, y0]
    """
    n1, n2, n3 = tgt.shape
    # 0. information of ref and tgt
    xyr = xyicr[:, :2]
    icr_r = xyicr[:, 2:]
    xyt = xyict[:, :2]
    ict_t = xyict[:, 2:]

    # 1. convert xyt to ict_r (ic of tgt in ref geometry)
    ict_r = apply_transform(xyt, transr, True)

    # 2. generate gridt_r based on ict_r  (grid of tgt in ref geometry)
    start1, end1 = round(ict_r[:, 0].min()), round(ict_r[:, 0].max())
    start2, end2 = round(ict_r[:, 1].min()), round(ict_r[:, 1].max())
    X, Y = np.meshgrid(np.arange(start1, end1 + 1),
                       np.arange(start2, end2 + 1),
                       indexing='ij')
    gridt_r = np.c_[X.flatten(), Y.flatten()]

    # we need know the start point of tgt compared to ref (assume ref start at 0, 0)
    shift_x = gridt_r[:, 0].min() - icr_r[:, 0].min()
    shift_y = gridt_r[:, 1].min() - icr_r[:, 1].min()

    # using kd-tree to filter the grid points (avoid interpolation on the blank area)
    tree = cKDTree(ict_r)
    distances, indices = tree.query(gridt_r, k=1)
    mask = distances <= 1  # using distance to filter the grid points
    to_interp = gridt_r[mask]

    # 3. convert gridt_r to xy_gt
    xy_gt = apply_transform(to_interp, transr, False)

    # 4. convert xy_gt to ic_gt_t in tgt geometry
    ic_gt_t = apply_transform(xy_gt, transt, True)

    # 5. interpolate data_gt_t using datat
    ic_gt_t[:, 0] -= ict_t[:, 0].min()
    ic_gt_t[:, 1] -= ict_t[:, 1].min()

    # another mask to filter the grid points, avoid interpolation out of the boundary
    mask2 = (ic_gt_t[:, 0] >= 0) & (ic_gt_t[:, 0] < n1-1) & \
            (ic_gt_t[:, 1] >= 0) & (ic_gt_t[:, 1] < n2-1)
    ic_gt_t = ic_gt_t[mask2]
    to_interp = to_interp[mask2]

    # interpolate data_gt_t using datat
    tgt_interp = interp3d(tgt, ic_gt_t, order)

    # 6. reshape data_gt_t to meet the 3d shape
    # the shape of output is (nx, ny, n3)
    nx = end1 - start1 + 1
    ny = end2 - start2 + 1

    output = np.zeros((nx, ny, n3), dtype=tgt.dtype)

    ix = (to_interp[:, 0] - start1).astype(int)
    iy = (to_interp[:, 1] - start2).astype(int)

    output[ix, iy, :] = tgt_interp

    return output, [shift_x, shift_y]


@njit
def interp3d(volume: np.ndarray, points: np.ndarray, order: int = 1):
    """
    3D body data is interpolated at a given 2D point.

    Parameters
    -----------
    volume : np.ndarray
        shape is (n1, n2, n3)
    points : np.ndarray
        shape is (N, 2), each row is [x, y]
    order : int
        order of interpolate, can be one of 1, 2, 3。

    Returns
    --------
    out : np.ndarray
        shape is (N, n3)
    """
    n1, n2, n3 = volume.shape
    N = points.shape[0]
    out = np.zeros((N, n3), dtype=volume.dtype)

    for idx in range(N):
        x = points[idx, 0]
        y = points[idx, 1]

        # Ensure that x and y are within the valid range
        # This also means the boundary is interpolated by the repeated values
        x = min(max(x, 0.0), n1 - 1.000001)
        y = min(max(y, 0.0), n2 - 1.000001)

        if order == 1:
            x0 = int(np.floor(x))
            y0 = int(np.floor(y))
            x1 = x0 + 1
            y1 = y0 + 1

            x1 = min(x1, n1 - 1)
            y1 = min(y1, n2 - 1)

            x_frac = x - x0
            y_frac = y - y0

            w00 = (1 - x_frac) * (1 - y_frac)
            w01 = (1 - x_frac) * y_frac
            w10 = x_frac * (1 - y_frac)
            w11 = x_frac * y_frac

            v00 = volume[x0, y0, :]
            v01 = volume[x0, y1, :]
            v10 = volume[x1, y0, :]
            v11 = volume[x1, y1, :]

            out[idx, :] = w00 * v00 + w01 * v01 + w10 * v10 + w11 * v11

        elif order == 2:
            # obtain the 9 grid points around the target point
            x_indices = np.array(
                [int(np.floor(x) - 0.5) + i for i in range(3)])
            y_indices = np.array(
                [int(np.floor(y) - 0.5) + i for i in range(3)])

            # Ensure that the indices are within the valid range
            x_indices = np.clip(x_indices, 0, n1 - 1)
            y_indices = np.clip(y_indices, 0, n2 - 1)

            # Obtain the node coordinates
            x_nodes = x_indices.astype(np.float64)
            y_nodes = y_indices.astype(np.float64)

            # Calculate the Lagrange basis functions
            Lx = np.zeros(3)
            Ly = np.zeros(3)

            for i in range(3):
                xi = x_nodes[i]
                Li = 1.0
                for j in range(3):
                    if i != j:
                        xj = x_nodes[j]
                        Li *= (x - xj) / (xi - xj + 1e-10)
                Lx[i] = Li

            for i in range(3):
                yi = y_nodes[i]
                Li = 1.0
                for j in range(3):
                    if i != j:
                        yj = y_nodes[j]
                        Li *= (y - yj) / (yi - yj + 1e-10)
                Ly[i] = Li

            # Calculate the interpolated value
            value = np.zeros(n3, dtype=volume.dtype)
            for i in range(3):
                xi = x_indices[i]
                for j in range(3):
                    yj = y_indices[j]
                    w = Lx[i] * Ly[j]
                    v = volume[xi, yj, :]
                    value += w * v
            out[idx, :] = value

        elif order == 3:
            # obtain the 16 grid points around the target point
            x_indices = np.array([int(np.floor(x) - 1) + i for i in range(4)])
            y_indices = np.array([int(np.floor(y) - 1) + i for i in range(4)])

            # Ensure that the indices are within the valid range
            x_indices = np.clip(x_indices, 0, n1 - 1)
            y_indices = np.clip(y_indices, 0, n2 - 1)

            # Obtain the node coordinates
            x_nodes = x_indices.astype(np.float64)
            y_nodes = y_indices.astype(np.float64)

            # Calculate the Lagrange basis functions
            Lx = np.zeros(4)
            Ly = np.zeros(4)

            for i in range(4):
                xi = x_nodes[i]
                Li = 1.0
                for j in range(4):
                    if i != j:
                        xj = x_nodes[j]
                        Li *= (x - xj) / (xi - xj + 1e-10)
                Lx[i] = Li

            for i in range(4):
                yi = y_nodes[i]
                Li = 1.0
                for j in range(4):
                    if i != j:
                        yj = y_nodes[j]
                        Li *= (y - yj) / (yi - yj + 1e-10)
                Ly[i] = Li

            # calculate the interpolated value
            value = np.zeros(n3, dtype=volume.dtype)
            for i in range(4):
                xi = x_indices[i]
                for j in range(4):
                    yj = y_indices[j]
                    w = Lx[i] * Ly[j]
                    v = volume[xi, yj, :]
                    value += w * v
            out[idx, :] = value

        else:
            raise ValueError("Order must be 1, 2, or 3")

    return out


def interp(p: np.ndarray, d: np.ndarray) -> np.ndarray:
    """
    Interpolates a 1d series in 3d space by **linear** interpolation.
    
    Parameters
    ----------
    p : ArrayLike
        (N, 2), each row is [x, y], normalized, i.e., 0 <= x, y <= 1
    d : ArrayLike
        shape is (N, 4, n3)
    """
    x = p[:, 0]
    y = p[:, 1]

    # linear interpolation
    # fmt: off
    out = (d[:, 0, :] * (1 - x)[:, None] * (1 - y)[:, None] +
           d[:, 1, :] * (1 - x)[:, None] * y[:, None] +
           d[:, 2, :] * x[:, None] * (1 - y)[:, None] +
           d[:, 3, :] * x[:, None] * y[:, None])

    # fmt: on

    return out


def interpolate_path(points, di=1):
    """
    Interpolates a path from the given points with a given step size.
    """
    points = np.array(points)

    diffs = np.diff(points, axis=0)
    distances = np.sqrt((diffs**2).sum(axis=1))
    cum_distances = np.concatenate(([0], np.cumsum(distances)))
    total_distance = cum_distances[-1]

    distances_interp = np.arange(0, total_distance + di, di)
    x_points = np.interp(distances_interp, cum_distances, points[:, 0])
    y_points = np.interp(distances_interp, cum_distances, points[:, 1])

    indices = np.searchsorted(distances_interp, cum_distances)

    return np.column_stack((x_points, y_points)), indices


def extract_data(data, p):
    """
    Parameters
    ----------
    segy: str
        segyfile
    p : ArrayLike
        shape is (N, 2)
    """
    n1, n2, n3 = data.shape

    N = p.shape[0]
    xq = p[:, 0]
    yq = p[:, 1]

    i = np.clip(np.searchsorted(np.arange(n1), xq, side="right"), 1, n1 - 1)
    j = np.clip(np.searchsorted(np.arange(n2), yq, side="right"), 1, n2 - 1)

    x0 = i - 1
    x1 = i
    y0 = j - 1
    y1 = j

    # normalized path points
    pout = np.zeros_like(p, np.float32)
    pout[:, 0] = p[:, 0] - i + 1
    pout[:, 1] = p[:, 1] - j + 1

    assert pout.min() >= 0 and pout.max() <= 1

    grid_points = np.stack(
        [
            np.array([x0, y0]).T,
            np.array([x0, y1]).T,
            np.array([x1, y0]).T,
            np.array([x1, y1]).T
        ],
        axis=1,
    )

    all_points = grid_points.reshape(-1, 2)
    unique, inverse = np.unique(all_points, axis=0, return_inverse=True)
    unique_data = np.array([data[i, j] for i, j in unique])

    return pout, unique_data[inverse].reshape(N, 4, n3)


def arbitray_line(data, points, di: float = 1):
    """
    data : SegyNP or np.ndarray
    points : the points to interpolate a path
    di : step
    """
    p, indices = interpolate_path(points, di)
    pout, pdata = extract_data(data, p)
    out = interp(pout, pdata)

    return out, p, indices



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.


def fromfile_ignore_header(*args, **kwargs):
    raise RuntimeError(
        "`fromfile_ignore_header` is deprecated and removed. Please use `collect` instead."
        " e.g., `d = cigsegy.collect(segy_name).reshape(ni, nx, nt)`")


def tofile_ignore_header(*args, **kwargs):
    raise RuntimeError(
        "`tofile_ignore_header` is deprecated and removed. Please use `collect` instead."
        " e.g., `d = cigsegy.collect(segy_name).reshape(ni, nx, nt)`")



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.

from datetime import datetime

import numpy as np
from .constinfo import kDataSampleFormatHelp
from cigsegy.cpp._CXX_SEGY import create_segy
from cigsegy import utils


def assemble_metainfo(shape, dformat=5, start=None, interval=None):
    ndim = len(shape)
    assert ndim >= 2 and ndim <= 4
    meta = {}
    meta['ndim'] = ndim
    meta['unit'] = 'm'
    meta['dformat'] = dformat
    # shape
    meta['nt'] = shape[-1]
    if ndim == 2:
        meta['ntrace'] = shape[0]
    else:
        meta['ni'] = shape[0]
        meta['nx'] = shape[1]
        if ndim == 4:
            meta['no'] = shape[2]
        meta['ntarce'] = np.prod(shape[:-1])

    # start
    if ndim == 2:
        if start is None:
            meta['start_time'] = 0
        else:
            assert isinstance(start, (int, np.integer))
            meta['start_time'] = start
    else:
        if start is None:
            start = [2000] * (len(shape) - 1)
            start.append(0)
        assert len(start) == ndim
        meta['start_iline'] = start[0]
        meta['end_iline'] = start[0] + meta['ni'] - 1
        meta['start_xline'] = start[1]
        meta['end_xline'] = start[1] + meta['nx'] - 1
        meta['start_time'] = start[-1]
        if ndim == 4:
            meta['start_offset'] = start[1]
            meta['end_offset'] = start[1] + meta['nx'] - 1

    # interval
    if ndim == 2:
        if interval is None:
            meta['dt'] = 2000
        else:
            assert isinstance(interval, (int, np.integer))
            meta['dt'] = interval
    else:
        if interval is None:
            interval = [25, 25, 2000]
        assert len(interval) == 3
        meta['di'] = interval[0]
        meta['dx'] = interval[1]
        meta['dt'] = interval[2]

    meta['istep'] = 1
    meta['xstep'] = 1
    meta['ostep'] = 1
    meta['scalar'] = 1

    return meta


def generate_textual(meta, textual=None):
    if textual is None:
        textual = parser_textual(meta)
    else:
        if isinstance(textual, list):
            if len(textual) <= 8:
                textual = parser_textual(meta, textual)
            else:
                textual = _parser_textual_list(textual)
        else:
            if len(textual) < 76 * 8:
                textual = parser_textual(meta, textual)
    assert len(textual) == 3200
    textual_header = utils.ascii_to_ebcdic(textual)
    return textual_header


def parser_textual(meta, custom=None):
    """
    """
    t = _gettime()
    ndim = meta['ndim']
    end_time = meta['start_time'] + (meta['nt'] - 1) * meta['dt'] // 1000
    timer = f"t: {meta['start_time']} - {end_time} ms"
    if ndim == 2:
        shape = f"n-trace: {meta['ntrace']}, n-time: {meta['nt']}"
        inter = f"dt = {meta['dt']//1000} ms"
        rge1 = timer
        rge2 = ""
        steps = ""
    elif ndim == 3:
        shape = f"n-iline: {meta['ni']}, n-xline: {meta['nx']}, n-time: {meta['nt']}"
        inter = f"di(iline) = {meta['di']:.2f} {meta['unit']}, dx(xline) = {meta['dx']:.2f} {meta['unit']}, dt = {meta['dt']//1000} ms"
        rge1 = f"inline: {meta['start_iline']} - {meta['end_iline']}, crossline: {meta['start_xline']} - {meta['end_xline']}"
        rge2 = timer
        steps = f"iline: {meta['istep']:3}, xline: {meta['xstep']:3}"
    elif ndim == 4:
        shape = f"n-iline: {meta['ni']}, n-xline: {meta['nx']}, n-offset: {meta['no']}, n-time: {meta['nt']}"
        inter = f"di(iline) = {meta['di']:.2f} {meta['unit']}, dx(xline) = {meta['dx']:.2f} {meta['unit']}, dt = {meta['dt']//1000} ms"
        rge1 = f"inline: {meta['start_iline']} - {meta['end_iline']}, crossline: {meta['start_xline']} - {meta['end_xline']}"
        rge2 = f"offset: {meta['start_offset']} - {meta['end_offset']}, " + timer
        steps = f"iline: {meta['istep']:3}, xline: {meta['xstep']:3}, offset: {meta['ostep']:3}"

    dformat = f"{meta['dformat']} -- " + kDataSampleFormatHelp[meta['dformat']]

    # fmt: off
    if custom is None:
        custom = [
            f"C01 Create By CIGSEGY software (CIG, USTC, 2024), see:",
            f"C02 github: https://github.com/JintaoLee-Roger/cigsegy",
            f"C03 ",
            f"C04 Type: {ndim}D seismic  Created Time: {t}",
            f"C05 ",
            f"C06 ",
            f"C07 ",
            f"C08 ",
        ]
    else:
        if isinstance(custom, list):
            assert len(custom) <= 8
            if len(custom) < 8:
                custom += [''] * (8 - len(custom))
            custom = [f'C{i+1:02} ' + s[:76] for i, s in enumerate(custom)]
        else:
            out = []
            for i in range(8):
                start = i * 76
                if i * 76 >= len(custom):
                    out.append(f'C{i+1:02}')
                else:
                    end = (i+1) * 76
                    end = end if end < len(custom) else len(custom)
                    out.append(f'C{i+1:02} ' + custom[start:end])
            custom = out

    metainfo = [
        f"C09 ",
        f"C10 ",
        f"C11 Shape: {shape}",
        f"C12 Interval: {inter}",
        f"C13 Ranges: {rge1}",
        f"C14         {rge2}",
        f"C15 Line steps: {steps}",
        f"C16 ",
        f"C17 scalar to coordinates: {meta['scalar']}",
        f"C18 Data sample format: {dformat}",
        f"C19 ",
        f"C20 ",
        f"C21 ",
        f"C22 ",
        f"C23",
    ]
    keyloc = [
        f"C24 Binary header locations:",
        f"C25 Sample interval             : bytes 17",
        f"C26 Number of samples per trace : bytes 21",
        f"C27 Data sample format code     : bytes 25",
        f"C28 ",
        f"C29 ",
        f"C30 Trace header locations:",
        f"C31 Inline number               : bytes 189",
        f"C32 Crossline number            : bytes 193",
        f"C33 X coordinate                : bytes 181",
        f"C34 Y coordinate                : bytes 185",
        f"C35 Trace start time/depth      : bytes 105",
        f"C36 Offset                      : bytes 37",
        f"C37 ",
        f"C38 ",
        f"C39 ",
        f"C40 ",
    ]

    textual = custom + metainfo + keyloc

    if len(textual) != 40:
        raise ValueError(f"Header must contain exactly 40 lines, got {len(textual)}")

    textual = ''.join(f'{line.ljust(80)}' for line in textual)

    if len(textual) != 3200:
        raise ValueError(f"Header length must be exactly 3200 bytes, got {len(textual)}")

    return textual
    # fmt: on


class SegyCreate:
    """
    Create a SEG-Y file from an array.

    Parameters
    -----------
    fname : str
        out SEG-Y file name
    data : np.ndarray
        array
    metainfo : None
        meta information

    Examples
    ---------
    ```python
    # data.shape == (1101, 801, 1201)
    segyc = SegyCreate('out.segy', data)

    # 1. default
    segyc.create()

    # 2. set start_time, dt
    segyc.set_dt(4000) # 4 ms
    segyc.set_start_time(200) # 200 ms
    segyc.create()

    # 3. manually set iline, xline for each trace
    iline = np.arange(2000, 2000+data.shape[0])
    xline = np.arange(3000, 3000+data.shape[1])
    il, xl = np.meshgrid(iline, xline, indexing='ij')
    keys = np.c_[il.flatten(), xl.flatten()]
    segyc.create(keys)

    # 4. manually set iline, xline, x, y for each trace
    iline = np.arange(2000, 2000+data.shape[0])
    xline = np.arange(3000, 3000+data.shape[1])
    x = np.arange(605959, 605959 + data.shape[0]*25)
    y = np.arange(6058307, 6058307 + data.shape[1]*12)
    il, xl = np.meshgrid(iline, xline, indexing='ij')
    x, y = np.meshgrid(x, y, indexing='ij')
    keys = np.c_[il.flatten(), xl.flatten(), x.flatten(), y.flatten()]
    segyc.create(keys)
    ```
    """

    def __init__(self, fname: str, data: np.ndarray, metainfo=None):
        self.filename = fname
        self.textual_header = None
        self.binary_header = None
        self.trace_header = None
        assert data.ndim >= 2 and data.ndim <= 4
        self.data = data
        self.shape = data.shape
        if metainfo is None:
            metainfo = assemble_metainfo(data.shape)
        self.metainfo = metainfo
        self._ndim = len(self.shape)

    def set_textual_header(self, textual=None):
        self.textual_header = generate_textual(self.metainfo, textual)

    def set_binary_header(self, bheader: np.ndarray = None):
        if bheader is None:
            bheader = self._init_bheader()
        bheader = bheader.view(np.uint8)
        assert bheader.size == 400
        self._set_keyi2(bheader, 17, self.metainfo['dt'])
        self._set_keyi2(bheader, 21, self.metainfo['nt'])
        self._set_keyi2(bheader, 25, self.metainfo['dformat'])
        self.binary_header = bheader

    def set_trace_header(self, theader: np.ndarray = None):
        if theader is None:
            theader = self._init_theader()
        theader = theader.view(np.uint8)
        assert theader.size == 240
        self._set_keyi2(theader, 69, self.metainfo['scalar'])
        self._set_keyi2(theader, 115, self.metainfo['nt'])
        self._set_keyi2(theader, 117, self.metainfo['dt'])
        self._set_keyi2(theader, 105, self.metainfo['start_time'])
        self.trace_header = theader

    def set_dt(self, dt):
        self.metainfo['dt'] = dt
        if self.binary_header is not None:
            self._set_keyi2(self.binary_header, 17, dt)
            self._set_keyi2(self.trace_header, 117, dt)

    def set_start_time(self, start_time):
        self.metainfo['start_time'] = start_time
        if self.trace_header is not None:
            self._set_keyi2(self.trace_header, 105, start_time)

    def copy_textual_from(self, segyname):
        with open(segyname, 'r') as f:
            self.textual_header = str(f.read(3200))

    def copy_bheader_from(self, segyname):
        self.binary_header = np.fromfile(
            segyname,
            dtype=np.uint8,
            count=400,
            offset=3200,
        )
        self._set_keyi2(self.binary_header, 21, self.shape[-1])

    def copy_theader_from(self, segyname):
        self.trace_header = np.fromfile(
            segyname,
            dtype=np.uint8,
            count=240,
            offset=3600,
        )
        self._set_keyi2(self.trace_header, 115, self.shape[-1])

    def create(self, keys: np.ndarray = None):
        if keys is None:
            keys = self._generate_keys()
        assert keys.ndim == 2
        if (self._ndim < 4 and keys.shape[1] == 2) or (self._ndim == 4
                                                       and keys.shape[1] == 3):
            xy = self._generate_xy()
            assert xy.shape[0] == keys.shape[0]
            keys = np.concatenate([keys, xy], axis=1)
        if self._ndim < 4:
            assert keys.shape[1] == 4
        else:
            assert keys.shape[1] == 5

        assert keys.shape[0] == np.prod(self.shape[:-1])

        if self.textual_header is None:
            self.set_textual_header()
        if self.binary_header is None:
            self.set_binary_header()
        if self.trace_header is None:
            self.set_trace_header()

        create_segy(self.fname, self.data, keys, self.textual_header,
                    self.binary_header, self.trace_header)

    # fmt: off
    def _generate_keys(self):
        if self._ndim == 2:
            x, y = np.meshgrid(np.array([10]), np.arange(100, self.shape[0]+100), indexing='ij')
            keys = np.c_[x.flatten(), y.flatten()]
        else:
            ib = self.metainfo['start_iline']
            ie = ib + self.shape[0]
            xb = self.metainfo['start_xline']
            xe = xb + self.shape[1]
            if self._ndim == 3:
                x, y = np.meshgrid(np.arange(ib, ie, 1), np.arange(xb, xe, 1), indexing='ij')
                keys = np.c_[x.flatten(), y.flatten()]
            else:
                ob = self.metainfo['start_offset']
                oe = ob + self.shape[2]
                x, y, z = np.meshgrid(np.arange(ib, ie, 1), np.arange(xb, xe, 1), np.arange(ob, oe, 1), indexing='ij')
                keys = np.c_[x.flatten(), y.flatten(), z.flatten()]
        return keys

    def _generate_xy(self):
        if self._ndim == 2:
            x, y = np.meshgrid(np.array([]), np.arange(), indexing='ij')
            xy = np.c_[x.flatten(), y.flatten()]
        else:
            di = self.metainfo['di']
            dx = self.metainfo['dx']
            ni = np.arange(108232, 108232+di*self.shape[0], di)
            nx = np.arange(413897, 413897+dx*self.shape[1], dx)
            if self._ndim == 3:
                x, y = np.meshgrid(x, y, indexing='ij')
                xy = np.c_[x.flatten(), y.flatten()]
            else:
                x, y, z = np.meshgrid(x, y, np.arange(10, self.shape[2]+10), indexing='ij')
                xy = np.c_[x.flatten(), y.flatten(), z.flatten()]

        return xy
    # fmt: on

    def _set_keyi2(self, header: np.ndarray, loc: int, value: int):
        loc = loc - 1
        header[loc:loc + 2] = np.array([value], dtype='>i2').view(np.uint8)
        return header

    def _set_keyi4(self, header: np.ndarray, loc: int, value: int):
        loc = loc - 1
        header[loc:loc + 4] = np.array([value], dtype='>i4').view(np.uint8)
        return header

    def _get_keyi2(self, header: np.array, loc: int):
        loc = loc - 1
        return np.frombuffer(header[loc:loc + 2], dtype='>i2')[0]

    def _get_keyi4(self, header: np.array, loc: int):
        loc = loc - 1
        return np.frombuffer(header[loc:loc + 4], dtype='>i4')[0]

    def _init_bheader(self):
        out = np.zeros((400, ), np.uint8)
        self._set_keyi4(out, 5, 1000)
        self._set_keyi2(out, 13, 1)
        self._set_keyi2(out, 55, 1)
        return out

    def _init_theader(self):
        out = np.zeros((240, ), np.uint8)
        if self._ndim == 2:
            self._set_keyi2(out, 29, 0)
        elif self._ndim == 3:
            self._set_keyi2(out, 29, 4)
        else:
            self._set_keyi2(out, 29, 2)

        return out


def _gettime():
    return datetime.now().strftime("%Y-%m-%d %H:%M")


def _parser_textual_list(text):
    textual = [f'C{i:02} ' + t for i, t in enumerate(text)]
    for i in range(len(text), 40):
        textual.append(f'C{i:02} ')

    textual = ''.join(f'{line.ljust(80)}' for line in textual)
    return textual



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.

from pathlib import Path
from typing import List
import warnings
import numpy as np
# from cigsegy import get_trace_keys
from cigsegy.cpp._CXX_SEGY import Pysegy
from .constinfo import *


def ebcdic_to_ascii(ebcdic_str: bytes) -> str:
    """Convert EBCDIC encoded string to ASCII."""
    return ''.join(kEBCDICtoASCIImap.get(byte, '?') for byte in ebcdic_str)


def ascii_to_ebcdic(ascii_str: str) -> bytes:
    """Convert ASCII encoded string to EBCDIC."""
    return bytes(kASCIItoEBCDICmap.get(char, 0) for char in ascii_str)


def parse_bheader(bheader: np.ndarray):
    """
    Parse binary header.
    """
    assert bheader.size == 400, "Binary header size must be 400 bytes."
    out, hstring = _parse_header(bheader, kBinaryHeaderHelp)

    return out, hstring


def parse_theader(theader: np.ndarray):
    """
    Parse trace header.
    """
    assert theader.size == 240, "Trace header size must be 240 bytes."
    out, hstring = _parse_header(theader, kTraceHeaderHelp)

    return out, hstring


def eval_iline(segy: Pysegy) -> int:
    """
    To guess the inline location of the segy

    Parameters
    ----------
    segy : Pysegy
        input Pysegy class

    Returns
    -------
    List
        possible result
    """
    ntrace = segy.ntrace
    options = [189, 5, 9, 221, 13, 17]
    select = [189, 5, 9, 221, 13, 17]

    for op in options:
        l0 = _get_keys4(segy, op, 0)
        ll = _get_keys4(segy, op, ntrace - 1)
        l2 = _get_keys4(segy, op, ntrace // 2)
        # line number should  not be negative
        if sum([x >= 0 for x in [l0, ll, l2]]) != 3:
            select.remove(op)
            continue
        # line number should be increasing/decreasing
        if l0 == ll or l0 == l2 or ll == l2:
            select.remove(op)
            continue
        # ni is too large
        if max([l0, ll, l2]) - min([l0, ll, l2]) > min(ntrace // 10 - 1, 50000): # yapf: disable
            select.remove(op)
            continue
        # line number should be increasing/decreasing
        if (l0 < l2 and l2 > ll) or (l0 > l2 and l2 < ll):
            select.remove(op)
            continue
        ls = _get_keys4(segy, op, ntrace // 2, ntrace // 2 + 50)
        if len(np.unique(ls)) > 10:
            select.remove(op)
            continue
        return op

    raise RuntimeError("Cannot evaluate inline location")


def is_valid_arr(arr, z_eval=False):
    arr_tuples = [tuple(row) for row in arr]
    if len(arr_tuples) != len(set(arr_tuples)):
        return False

    if z_eval:
        arr = arr[:, :2]
        z = arr[:, 2]
    x = arr[:, 0]
    y = arr[:, 1]

    if z_eval:
        ic = (z, y, x)
        dc = (-z, -y, -x)
    else:
        ic = (y, x)
        dc = (-y, -x)

    idx_asc = np.lexsort(ic)
    arr_asc = arr[idx_asc]

    idx_desc = np.lexsort(dc)
    arr_desc = arr[idx_desc]

    if np.array_equal(arr, arr_asc):
        return True
    elif np.array_equal(arr, arr_desc):
        return True
    else:
        return False


def guess2(segy_name: str,
           iline=None,
           xline=None,
           offset=None,
           istep=None,
           xstep=None,
           ostep=None,
           xloc=None,
           yloc=None):
    """
    guess the locations and steps of inline and crossline
    """
    if isinstance(segy_name, Pysegy):
        segy = segy_name
    else:
        segy = Pysegy(str(segy_name))

    N = segy.ntrace

    # 0. evaluate offset
    offset = 37 if offset is None else offset
    oostep, is4d = eval_offset(segy, offset)
    if ostep is None:
        oostep = ostep
    elif abs(oostep) < abs(ostep):
        warnings.warn(f"We evaluate istep == {istepx}, but you give istep == {istep}") # yapf: disable
    elif oostep * ostep < 0:
        warnings.warn(f"We evaluate istep == {istepx}, but you give istep == {istep}, order maybe wrong") # yapf: disable


    # 1. evaluate iline
    iline = eval_iline(segy) if iline is None else iline
    ni = _get_keys4(segy, iline, N - 1) - _get_keys4(segy, iline, 0)

    # 2. evaluate istep
    jumpi = N // ni // 4 if ni != 0 else 0
    if istep is None and ni == 0:
        istep = 1
    else:
        start = N // 4
        ils = [_get_keys4(segy, iline, start)]
        idx = [start]
        start += jumpi
        while len(ils) < 6 and start < N:
            il = _get_keys4(segy, iline, start)
            start += jumpi
            if il != idx[-1]:
                idx.append(start)
                ils.append(il)
        ilset = set(ils)
        ilset.update(_get_keys4(segy, iline, N // 4, N // 4 + 10).flatten())

        dif = np.diff(np.array(sorted(ilset)))
        istepx = dif.min() if dif[0] > 0 else dif.max()

        if istep is None:
            istep = istepx
        elif abs(istepx) < abs(istep):
            warnings.warn(f"We evaluate istep == {istepx}, but you give istep == {istep}") # yapf: disable
        elif istepx * istep < 0:
            warnings.warn(f"We evaluate istep == {istepx}, but you give istep == {istep}, order maybe wrong") # yapf: disable

    ni = ni / istep + 1
    if ni < 0 or not (isinstance(ni, int) or (isinstance(ni, float) and ni.is_integer())): # yapf: disable
        raise RuntimeError(f"Cannot evaluate iline/istep, becuase ni = {ni}")

    # 3. evaluate xline
    di = None
    if xline is None:
        candidate = [193, 17, 21, 13]
        start, end = N // 2, N // 2 + 50
        keys = candidate + [iline, offset]
        keys = keys if is4d else keys[:-1] # need offset?
        d = _get_keys4(segy, keys, start, end)
        for i, cxl in enumerate(candidate):
            mnx = min(N - 1, N // ni * 6)
            m1 = _get_keys4(segy, cxl, N - 1)
            m2 = _get_keys4(segy, cxl, 0)
            if abs(m1 - m2) > mnx:
                continue

            di = d[:, [4, i, 5]]
            while np.all(di[:, 0] == di[0, 0]) and np.all(di[:, 1] == di[0, 1]) and end < N // 2 + 1100: # yapf: disable
                start += 50
                end += 50
                d_new = _get_keys4(segy, [xline, candidate[i], offset], start, end) # yapf: disable
                di = np.concatenate([di, d_new])
            if end > 1000:
                continue

            if sum(di >= 0) != len(di):
                continue
            if not is_valid_arr(di):
                continue
            xline = cxl
            break

        if xline is None:
            raise RuntimeError("Evaluate xline error")

    # 4. evaluate xstep
    if xstep is None:
        if di is None:
            di = _get_keys4(segy, [iline, xline, offset], start, end)


    # 7. evaluate xloc, yloc
    if xloc is None or yloc is None:
        pass

    return iline, xline, offset, istep, xstep, ostep, xloc, yloc, is4d


def guess(segy_name: str,
          iline=None,
          xline=None,
          offset=None,
          istep=None,
          xstep=None,
          ostep=None,
          xloc=None,
          yloc=None) -> List:
    """
    guess the locations and steps of inline and crossline
    """
    if isinstance(segy_name, Pysegy):
        segy = segy_name
    else:
        segy = Pysegy(str(segy_name))

    N = segy.ntrace
    offset = 37 if offset is None else offset
    # ostep, is4d = eval_offset(segy, offset)
    is4d = False
    ostep = 1

    iline = eval_iline(segy) if iline is None else iline

    # read 3 lines
    start = int(N // 3)
    lines = set()
    oix = []
    xlines = [193, 17, 21, 13] if xline is None else [xline]
    while True and (start + 400) <= segy.ntrace:
        part = _get_keys4(segy, [offset, iline, *xlines], start, start + 400)
        oix.append(part)
        lines.update(np.unique(part[:, 1]))
        start += 400
        if len(lines) >= 3:
            break

    # eval istep
    lines = np.array(sorted(list(lines)))
    # print(lines)
    dif = np.diff(lines)
    istepx = dif.min() if dif[0] > 0 else dif.max()

    # line: n, n+1, n+2, extract the data of line n+1
    oix = np.concatenate(oix)
    idx = np.where(np.diff(oix[:, 1]) != 0)[0][:2] + 1
    oix = oix[idx[0]:idx[1], :]
    nig = oix.shape[0]

    def _double_check_istep():
        start2 = int(N // 3 * 2)
        oix2 = []
        lines2 = set()
        while True and (start2 + nig) <= segy.ntrace:
            part2 = _get_keys4(segy, iline, start2, start2 + nig)
            oix2.append(part2)
            lines2.update(np.unique(part2))
            start2 += nig
            if len(lines2) >= 3:
                break
        lines2 = np.array(sorted(list(lines2)))
        dif2 = np.diff(lines2)
        istepx2 = dif2.min() if dif2[0] > 0 else dif2.max()
        return istepx2

    if abs(istepx) > 1:  # if istep is not 1, double check
        istepx2 = _double_check_istep()
        if abs(istepx2) < abs(istepx):
            istepx = istepx2

    if istep is not None and istep != istepx:
        warnings.warn(f"You set istep={istep}, but we scan the istep is {istepx}, be careful!") # yapf: disable
    else:
        istep = istepx

    # eval xline
    def _eval_xline(i):
        dif = np.diff(oix[:, i])
        idx = np.where(dif != 0)[0]
        # values, counts = np.unique(idx, return_counts=True)
        dif = dif[dif != 0]
        if len(dif) == 0:
            return 1
        xstepi = dif.min() if dif[0] > 0 else dif.max()
        return xstepi

    if len(xlines) == 1:
        xstepx = _eval_xline(2)
        xidx = 2
    else:
        for i in range(len(xlines)):
            if abs(_get_keys4(segy, xlines[i], N - 1) - _get_keys4(segy, xlines[i], 0)) > N // 10: # yapf: disable
                continue
            xstepx = _eval_xline(i + 2)
            if abs(xstepx) > 0 and abs(xstepx) < 100:
                xline = xlines[i]
                xidx = i + 2
                # oix = oix[:, [0, 1, i + 2]]
                break

    if xstep is not None and xstep != xstepx:
        warnings.warn(f"You set xstep={xstep}, but we scan the xstep is {xstepx}, be careful!") # yapf: disable
    elif xstepx == 0:
        raise RuntimeError("Cannot evaluate xline location, we evaluate xstep is 0") # yapf: disable
    else:
        xstep = xstepx

    # eval offset
    change_points = np.where(np.diff(oix[:, xidx]) != 0)[0] + 1
    segments = np.split(oix[:, xidx], change_points)
    count = sum(1 for seg in segments if len(seg) > 1)
    if count > 15:
        ostep, is4d = eval_offset(segy, offset)

    # eval xloc and yloc
    if xloc is None or yloc is None:
        xys = _get_keys4(segy, [181, 185, 73, 77], 0, 100)
        if len(np.unique(xys[:, 2])) == 0 and len(np.unique(xys[:, 3])) == 0:
            xloc, yloc = 181, 185
        elif len(np.unique(xys[:, 0])) == 0 and len(np.unique(xys[:, 1])) == 0:
            xloc, yloc = 73, 77
        else:
            scalar = segy.keyi2(0, 71)
            if scalar < -1000 or scalar > 1000:
                scalar = 1
            scalar = 1 if scalar == 0 else scalar
            scalar = -1 / scalar if scalar < 0 else scalar
            xys = xys * scalar
            d1 = ((xys[-1, 0] - xys[0, 0])**2 +
                  (xys[-1, 1] - xys[0, 1])**2)**0.5
            d2 = ((xys[-1, 2] - xys[0, 2])**2 +
                  (xys[-1, 3] - xys[0, 3])**2)**0.5
            if d1 < 3 and d2 > 3:
                xloc, yloc = 73, 77
            else:
                xloc, yloc = 181, 185
    return iline, xline, offset, istep, xstep, ostep, xloc, yloc, is4d


def eval_offset(segyname, offset: int = 37) -> int:
    """
    return the ostep, if return -1, the file is not prestack SEG-Y
    """
    if isinstance(segyname, Pysegy):
        segy = segyname
    else:
        segy = Pysegy(str(segyname))

    ntrace = segy.ntrace
    ks = _get_keys4(segy, offset, ntrace // 3, ntrace // 3 + 200)

    if not isinstance(segyname, Pysegy):
        segy.close()

    if len(np.unique(ks)) == 1:
        return 0, False

    dif = np.diff(ks)
    udif, count = np.unique(dif, return_counts=True)
    if (len(udif) > 10):
        warnings.warn("offset is not constant, and is unsorted, cannot evaluate by `scan`. So we treat this file as a 3D.") # yapf: disable
        # raise RuntimeError("offset is not constant, and is unsorted, cannot evaluate by `scan`") # yapf: disable
        return 0, False

    ostep = udif[np.argmax(count)]
    if ostep == 0:
        return 0, False
    return ostep, True


def parse_metainfo(meta: dict):
    out = ""

    # shape information
    shapeinfo = "shape: "
    if meta['ndim'] == 2:
        shapeinfo += f"(n-trace, n-time) = ({meta['ntrace']}, {meta['nt']})"
    elif meta['ndim'] == 3:
        shapeinfo += f"(n-inline, n-crossline, n-time) = ({meta['ni']}, {meta['nx']}, {meta['nt']})"
    else:
        shapeinfo += f"(n-inline, n-crossline, n-offset, n-time) = ({meta['ni']}, {meta['nx']}, {meta['no']}, {meta['nt']})"

    out += shapeinfo + "\n"
    out += f"N traces: {meta['ntrace']}\n"

    # interval
    intervalinfo = "interval: "
    if meta['ndim'] == 2:
        intervalinfo += f"dt = {meta['dt']//1000} ms"
    else:
        intervalinfo += f"di(iline) = {meta['di']:.2f} {meta['unit']}, dx(xline) = {meta['dx']:.2f} {meta['unit']}, dt = {meta['dt']//1000} ms"
    out += intervalinfo + "\n"

    # range
    rangeinfo = "range: "
    end_time = meta['start_time'] + (meta['nt'] - 1) * meta['dt'] / 1000
    timer = f"t: {meta['start_time']} - {end_time} ms"
    if meta['ndim'] > 2:
        rangeinfo += f"inline: {meta['start_iline']} - {meta['end_iline']}, crossline: {meta['start_xline']} - {meta['end_xline']}, "
    if meta['ndim'] == 4:
        rangeinfo += f"offset: {meta['start_offset']} - {meta['end_offset']}, "
    rangeinfo += timer
    out += rangeinfo + "\n"

    tracesort = "trace sorting code: " + kTraceSortingHelp[meta['trace_sorting_code']] # yapf: disable
    out += tracesort + "\n"

    dformat = f"scalar: {meta['scalar']}, data format: " + kDataSampleFormatHelp[meta['dformat']] # yapf: disable
    out += dformat + "\n"

    kinfo = "(key info) "
    stepinfo = "           "
    if meta['ndim'] == 3:
        kinfo += f"iline: {meta['iline']:3}, xline: {meta['xline']:3}"
        stepinfo += f"istep: {meta['istep']:3}, xstep: {meta['xstep']:3}"
    if meta['ndim'] == 4:
        kinfo += f", offset: {meta['offset']:3}"
        stepinfo += f"ostep: {meta['ostep']:3}"
    kinfo += f", xloc: {meta['xloc']:3}, yloc: {meta['yloc']:3}\n"
    out += kinfo
    out += stepinfo + "\n"

    return out


def post_process_meta(segy: Pysegy, meta: dict, apply_scalar=True):
    unit = segy.bkeyi2(55)
    if apply_scalar:
        if meta['scalar'] == 0:
            meta['scalar'] = 1
        scalar = -1 / meta['scalar'] if meta['scalar'] < 0 else meta['scalar']
        meta['di'] *= scalar
        meta['dx'] *= scalar
    if unit == 2:
        meta['unit'] = 'ft'
    else:
        meta['unit'] = 'm'
    return meta


############# Internal functions #############


def _get_keys4(segy: Pysegy, keyloc, beg=-1, end=0):
    if beg < 0:
        beg = 0
        end = segy.ntrace
    if end < 0:
        end = segy.ntrace
    if end == 0:
        end = beg + 1

    if isinstance(keyloc, int):
        keyloc = [keyloc]
    d = segy.get_trace_keys(keyloc, [4] * len(keyloc), beg, end).squeeze()
    if d.size == 1 and d.ndim == 0:
        return int(d)
    elif d.size == 1 and d.ndim == 1:
        return int(d[0])
    else:
        return d


def _to_number(d, loc, ksize, dtype):
    return np.frombuffer(d[loc - 1:loc + ksize - 1].tobytes(), dtype=dtype)[0]


def _parse_header(header: np.ndarray, help_dict: dict) -> dict:
    """
    Parse binary header.
    """
    out = {}
    hstring = []
    for key, (disc, ksize) in help_dict.items():
        if ksize == 1:
            out[key] = header[key - 1]
        elif ksize == 2:
            out[key] = _to_number(header, key, ksize, '>i2')
        elif ksize == 4:
            out[key] = _to_number(header, key, ksize, '>i4')
        elif ksize == 8:
            out[key] = _to_number(header, key, ksize, '>i8')
        else:
            out[key] = 0
        hstring.append(f"{key:^3} - {key+ksize-1:^3}: {out[key]:<8} - {disc}")

    return out, hstring



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.
"""
_CXX_SEGY: A C++ core library for SEG-Y file operations, exposed to Python using pybind11.

This module provides high-performance, low-level functionality for handling SEG-Y files, a common format for geophysical data storage. The library is written in C++ for efficiency and is wrapped with pybind11 to be accessible from Python.

Key Features:
--------------
- **Scan SEG-Y Files**: Quickly scan SEG-Y files to extract metadata and other essential information without loading the entire file into memory.
  
- **Read SEG-Y Files**: Efficiently read seismic data and headers from SEG-Y files, allowing for fast data access.
  
- **Write SEG-Y Files**: Modify SEG-Y files with full control over file headers and format specifications (including header and data).
  
- **Create SEG-Y Files**: Initialize new SEG-Y files, specifying headers, formats, and data structures according to industry standards.

- **Efficient Functions**: data converting between IBM and IEEE floating-point formats.
"""

from typing import List, Tuple, overload

import numpy as np


class Pysegy:
    """
    Pysegy: A Python class for scanning, accessing, reading, and modifying SEG-Y file headers and data.

    The `Pysegy` class provides a high-level interface for working with SEG-Y files, which are widely used in the geophysical and seismic industries. 
    This class is implemented in C++ for performance reasons and is exposed to Python using pybind11, making it both powerful and easy to use.
    
    Example:
    --------
    ```python
    from cigsegy import Pysegy
    # from cigsegy.cpp._CXX_SEGY import Pysegy

    segy_file = Pysegy("example.segy")
    segy_file.setLocations(189, 193)
    segy_file.scan()
    # read the entire SEG-Y file
    d = segy_file.read()
    d.close()
    ```
    """

    def __init__(self, segy_name: str, write: bool = False) -> None:
        ...

    @property
    def ntrace():
        """
        number of traces
        """

    @property
    def nt():
        """
        nt
        """

    @property
    def ndim():
        """
        ndim
        """

    @property
    def shape():
        """
        ndim
        """

    @property
    def is_scanned():
        """
        is_scanned
        """

    def close() -> None:
        """
        close file
        """

    def show_progress(self, show: bool) -> None:
        """
        show progress
        """

    def setLocations(self, iline: int, xline: int, offset: int = 37) -> None:
        """ 
        set the inline, crossline and offset field of trace headers
        """

    def setInlineLocation(self, iline: int) -> None:
        """ 
        set the crossline field of trace headers (for reading segy)
        recommend: 189, 5, 9, default is 189.
        """

    def setCrosslineLocation(self, xline: int) -> None:
        """ 
        set the crossline field of trace headers (for reading segy)
        recommend: 193, 17, 21, default is 193.
        """

    def setOffsetLocation(self, offset: int) -> None:
        """ 
        set the offset field of trace headers (for reading segy)
        recommend: 37
        """

    def setXLocation(self, xloc: int) -> None:
        """
        set the x field of trace headers (for reading segy)
        recommend: 181, 73
        """

    def setYLocation(self, yloc: int) -> None:
        """
        set the y field of trace headers (for reading segy)
        recommend: 185, 77
        """

    def setXYLocations(self, xloc: int, yloc: int) -> None:
        """
        set the x and y field of trace headers (for reading segy)
        """

    def setInlineStep(self, istep: int) -> None:
        """ 
        set the step of inline (for reading segy)
        """

    def setCrosslineStep(self, xstep: int) -> None:
        """ 
        set the step of crossline (for reading segy)
        """

    def setOffsetStep(self, ostep: int) -> None:
        """ 
        set the step of offset (for reading segy)
        """

    def setSteps(self, istep: int, xstep: int, ostep: int = 1) -> None:
        """
        set inline, crossline and offset setps
        """

    def setFill(self, fill: float) -> None:
        """ 
        set a value for filling the missing trace (for reading segy),
        can be any float number or np.nan
        """

    def textual_header(self, coding='u') -> str:
        """
        obtain the 3200 bytes textual header.

        Parameters
        ---------
        coding : {'u', 'e', 'a'}, optional
            force the 3200 bytes in coding format, 'u' means guess by default,
            'e' means EBCDIC, 'a' means ASCII

        Returns
        -------
        str
            3200 bytes textual header
        """

    def bkeyi2(self, loc: int) -> int:
        """
        return the key value of the binary header in loc, 2 bytes
        """

    def bkeyi4(self, loc: int) -> int:
        """
        return the key value of the binary header in loc, 4 bytes
        """

    def keyi2(self, n: int, loc: int) -> int:
        """
        return the key value of the n-th trace header in loc, 2 bytes
        """

    def keyi4(self, n: int, loc: int) -> int:
        """
        return the key value of the n-th trace header in loc, 4 bytes
        """

    def iline(self, n: int) -> int:
        """
        return the inline value of the n-th trace
        """

    def xline(self, n: int) -> int:
        """
        return the crossline value of the n-th trace
        """

    def offset(self, n: int) -> int:
        """
        return the offset value of the n-th trace
        """

    def coordx(self, n: int) -> int:
        """
        return the X coordinate value of the n-th trace
        """

    def coordy(self, n: int) -> int:
        """
        return the Y coordinate value of the n-th trace
        """

    def get_binary_header(self) -> np.ndarray:
        """
        Obtain the 400 bytes binary header, return np.ndarray[np.uint8].

        Note: the binary header is raw data, i.e., big endian
        """

    def get_trace_header(self, n) -> np.ndarray:
        """
        Obtain the 240 bytes trace header of the n-th trace, return np.ndarray[np.uint8].

        Note: the trace header is raw data, i.e., big endian
        """

    def get_trace_keys(
        self,
        keys: List,
        length: List,
        beg: int,
        end: int,
    ) -> np.ndarray:
        """
        get the trace keys from beg to end

        Parameters
        ------------
        keys : List
            location
        length : List
            keys' length
        beg : int
            the start trace index
        end : int 
            the stop trace index (not included)

        Returns
        -------
        numpy.ndarray[numpy.int32]
            shape is (end-beg, len(keys))
        """

    def itrace(self, n: int) -> np.ndarray:
        """
        exract the data in the n-th trace
        """

    @overload
    def collect(self, beg: int, end: int, tbeg: int, tend: int) -> np.ndarray:
        """
        collect the data from the beg-th trace to the end-th trace (not included),
        time window is from tbeg to tend

        Parameters
        ----------
        beg : int
            the start trace index
        end : int
            the stop trace index (not included)
        tbeg : int
            the start time index
        tend : int
            the stop time index (not included)
        
        Returns
        -------
        numpy.ndarray
            shape is (end-beg, tend-tbeg)
        """

    @overload
    def collect(self, index: np.ndarray, tbeg: int, tend: int) -> np.ndarray:
        """
        Collect the data from the index array. This funcion is useful 
        when collect unsorted traces. if a index is negative, the trace 
        will be filled by fills (0 is default).

        Parameters
        -----------
        index : numpy.ndarray
            1D array, np.int32, 
        tbeg : int
            the start time index
        tend : int
            the stop time index (not included)
        
        Returns
        -------
        numpy.ndarray
            shape is (index.shape[0], tend-tbeg)

        Examples
        ---------
        >>> index = np.array([10, -1, 200, 300])
        >>> data = Pysegy('out.segy').collect(index, 0, 1000)
        >>> data.shape # (4, 1000)
        >>> np.allclose(data[1], 0) # True
        """

    def get_keylocs(self) -> dict:
        """
        get key locations in dict format
        """

    def get_metainfo(self) -> dict:
        """
        get metainfo in dict format.
        """

    def get_lineInfo(self) -> np.ndarray:
        """
        get lineinfo in np.ndarray format.
        if ndim == 3, each raw: 
            [iline, xstart, xend, trace_start, trace_end]
        if ndim == 4, each raw:
            [iline, xline, ostart, oend, trace_start, trace_end]


        Returns
        -------
        numpy.ndarray
            if ndim == 3, each raw: 
                [iline, xstart, xend, trace_start, trace_end]
            if ndim == 4, each raw:
                [iline, xline, ostart, oend, trace_start, trace_end]
        """

    def set_segy_type(self, ndim) -> None:
        """
        set the segy type, 2 or 3 or 4
        2 for a collect of traces,
        3 for poststack 3D volume,
        4 for prestack 4D volume/gather
        """

    def scan(self) -> None:
        """
        fast scan the file to obtain some meta infos
        """

    def read4d(
        self,
        ib: int,
        ie: int,
        xb: int,
        xe: int,
        ob: int,
        oe: int,
        tb: int,
        te: int,
    ) -> np.ndarray:
        """
        """

    def read3d(
        self,
        ib: int,
        ie: int,
        xb: int,
        xe: int,
        tb: int,
        te: int,
    ) -> np.ndarray:
        """
        """

    def read(self) -> np.ndarray:
        """
        """

    def read_tslice(self,
                    t: int,
                    stepi: int = 1,
                    stepx: int = 1) -> np.ndarray:
        """
        read time slice with stepi and stepx, which may faster than read()
        return a 2D array with shape: ((ni+stepi-1)//stepi, (nx+stepx-1)//stepx)
        """

    def tofile(
        self,
        binary_out_name: str,
        as_2d=False,
    ) -> None:
        """
        """

    def cut(self,
            outname: str,
            ranges: List,
            is2d: bool = False,
            textual: str = "") -> None:
        """
        """

    @overload
    def create_by_sharing_header(
        self,
        segy_name: str,
        src: np.ndarray,
        start: List,
        is2d: bool = False,
        textual: str = "",
    ) -> None:
        """
        """

    @overload
    def create_by_sharing_header(
        self,
        segy_name: str,
        src_file: str,
        shape: List,
        start: List,
        is2d: bool = False,
        textual: str = "",
    ) -> None:
        """
        """

    def set_bkeyi2(self, loc: int, val: int) -> None:
        """
        """

    def set_bkeyi4(self, loc: int, val: int) -> None:
        """
        """

    def set_keyi2(self, n: int, loc: int, val: int) -> None:
        """
        """

    def set_keyi4(self, n: int, loc: int, val: int) -> None:
        """
        """

    def set_iline(self, n: int, val: int) -> None:
        """
        """

    def set_xline(self, n: int, val: int) -> None:
        """
        """

    def set_offset(self, n: int, val: int) -> None:
        """
        """

    def set_coordx(self, n: int, val: int) -> None:
        """
        """

    def set_coordy(self, n: int, val: int) -> None:
        """
        """

    def write_itrace(self, data: np.ndarray, n: int) -> None:
        """
        """

    @overload
    def write_traces(
        self,
        data: np.ndarray,
        beg: int,
        end: int,
        tbeg: int,
        tend: int,
    ) -> None:
        """
        """

    @overload
    def write_traces(
        self,
        data: np.ndarray,
        index: np.ndarray,
        tbeg: int,
        tend: int,
    ) -> None:
        """
        """

    def write(self, data: np.ndarray) -> None:
        """
        """

    def write3d(
        self,
        data: np.ndarray,
        ib: int,
        ie: int,
        xb: int,
        xe: int,
        tb: int,
        te: int,
    ) -> None:
        """
        """

    def write4d(
        self,
        data: np.ndarray,
        ib: int,
        ie: int,
        xb: int,
        xe: int,
        ob: int,
        oe: int,
        tb: int,
        te: int,
    ) -> None:
        """
        """


def create_segy(
    segyname: str,
    src: np.ndarray,
    keys: np.ndarray,
    textual: str,
    bheader: np.ndarray,
    theader: np,
    ndarray,
) -> None:
    """
    create a new segy file from src data

    Parameters
    ----------
    segyname : str
        the name of the segy file
    src : np.ndarray, np.float32
        the source data in np.float32
    keys : np.ndarray, np.int32
        the keys of the traces, shape is (N, keysize), where N is the number of traces,
        and keysize can be 4, 5. N = src.shape[0]*...*src.shape[-2].
        if kesize is 4, each row contains [iline, xline, X, Y]
        if keysize is 5, each row contains [iline, xline, offset, X, Y]
    textual : str
        the textual header, 3200 bytes
    bheader : np.ndarray, np.uint8
        the binary header, 400 bytes
    theader : np.ndarray, np.uint8
        the trace header template, 240 bytes
    """


def ibm_to_ieee(value: float, is_big_endian: bool) -> float:
    """
    convert IBM floating point to IEEE floating point
    """


def ieee_to_ibm(value: float,
                is_litte_endian_input: bool,
                is_big_endian_output: bool = True) -> float:
    """
    convert IEEE floating point to IBM floating point
    """


def ibms_to_ieees(ibms: np.ndarray, is_big_endian: bool) -> np.ndarray:
    """
    convert IBM floating array to IEEE floating points
    """


def ieees_to_ibms(ieees: np.ndarray,
                  is_litte_endian_input: bool) -> np.ndarray:
    """
    convert IEEE floating array to IBM floating points
    """


def set_progress_callback(func: callable[[int, int], None]) -> None:
    """
    set the progress callback function

    Parameters
    ----------
    func : callable
        A callback function that takes two integer arguments: current and total.
    """


def set_global_show_progress(show: bool) -> None:
    """
    set the global progress bar visiable or not
    """



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.
"""
Plotting functions for visualizing survey coordinates, trace header keys, and other SEG-Y related data.

This module provides a collection of plotting functions tailored for geophysical data visualization. 
These functions are designed to help users graphically represent various aspects of SEG-Y files, 
such as survey coordinates and trace header keys. 

**The module may require additional dependencies for full functionality.**
"""

import numpy as np
from typing import Tuple
from cigsegy.cpp._CXX_SEGY import Pysegy
from cigsegy import ExceptionWrapper, tools
from cigsegy.interp import arbitray_line

try:
    import matplotlib.pyplot as plt
except BaseException as E:
    plt = ExceptionWrapper(E, "run `pip install matplotlib` to install the dependency") # yapf: disable


def plot_region(fname: str,
                mode: str = 'line',
                save: str = None,
                **kwargs) -> None:
    """
    plot the region map (x and y axis are inline and crossline)

    Parameters
    ----------
    segy : str or Pysegy
        input segy file
    loc : list
        contains 4 values, as [iline, xline, istep, xstep]
    save : str
        save to a png image
    """
    if mode not in ['line', 'cdpxy', 'xy']:
        raise RuntimeError(f"mode must be 'line' or 'cdpxy'/'xy', mode = {mode}") # yapf: disable

    geom = tools.get_lineInfo(fname, mode='geom', **kwargs)
    if mode == 'line':
        x, y = geom[:, 0], geom[:, 1]
    else:
        x, y = geom[:, 2], geom[:, 3]

    plt.fill(x, y, color=(0.9, 0.9, 0.9))
    plt.plot(x, y)

    k1, k2 = [x[0], y[0]], [x[-2], y[-2]]
    s = (x.max() - x.min()) / 50
    plt.arrow(k1[0], k1[1], k2[0]-k1[0], k2[1]-k1[1], head_width=s, head_length=s, ec='red', fc='red', zorder=4)

    plt.gca().invert_yaxis()
    # plt.gca().xaxis.set_ticks_position('top')

    plt.grid(True, linestyle='--')
    if mode == 'line':
        xlabel = f"Inline Number"
        ylabel = f"Crossline Number"
    else:
        xlabel = f"X"
        ylabel = f"Y"

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title('Region')
    plt.tight_layout()
    if save:
        plt.savefig(save, dpi=200, bbox_inches='tight', pad_inches=0.0)
    plt.show()


def plot_trace_keys(fname: str,
                    keyloc: int,
                    beg: int = 0,
                    end: int = 1000,
                    save: str = None) -> None:
    """
    plot the values (at keyloc in each trace) of the traces 
    range from beg to end .
    """
    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))
    assert beg >= 0 and end > beg + 1 and end <= segy.ntrace, "invalid beg and end"
    keys = segy.get_trace_keys([keyloc], [4], beg, end).squeeze()
    x = np.arange(beg, end)
    plt.plot(x, keys)
    plt.xlabel("Trace Number")
    plt.ylabel(f"Key values (at {keyloc})")
    plt.grid(True, linestyle='--')
    plt.tight_layout()
    if save:
        plt.savefig(save, dpi=200, bbox_inches='tight', pad_inches=0.0)
    plt.show()


def plot_trace_ix(fname: str,
                  iline: int,
                  xline: int,
                  beg: int = 0,
                  end: int = 1000,
                  figsize: Tuple = None,
                  save: str = None) -> None:
    """
    plot inline and crossline number of the traces range from
    beg to end 
    """
    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))
    assert beg >= 0 and end > beg + 1, "invalid beg and end"
    assert end < segy.ntrace, "end > trace_count"
    keys = segy.get_trace_keys([iline, xline], [4, 4], beg, end).squeeze()
    iv = keys[:, 0]
    xv = keys[:, 1]
    x = np.arange(beg, end)

    if figsize is None:
        figsize = (9, 4)
    plt.figure(figsize=figsize)
    plt.subplot(1, 2, 1)
    plt.plot(x, iv)
    plt.xlabel('Trace Number')
    plt.ylabel(f"Inline (at {iline})")
    plt.grid(True, linestyle='--')
    plt.subplot(1, 2, 2)
    plt.plot(x, xv)
    plt.xlabel('Trace Number')
    plt.ylabel(f"Crossline (at {xline})")
    plt.grid(True, linestyle='--')
    plt.tight_layout()
    if save:
        plt.savefig(save, dpi=200, bbox_inches='tight', pad_inches=0.0)
    plt.show()


def plot_trace_ixo(fname: str,
                   iline: int,
                   xline: int,
                   offset: int = 37,
                   beg: int = 0,
                   end: int = 1000,
                   figsize: Tuple = None,
                   save: str = None) -> None:
    """
    plot inline, crossline and offset number of the traces range from
    beg to end 
    """
    if isinstance(fname, Pysegy):
        segy = fname
    else:
        segy = Pysegy(str(fname))
    assert beg >= 0 and end > beg + 1, "invalid beg and end"
    assert end < segy.ntrace, "end > trace_count"
    keys = segy.get_trace_keys([iline, xline, offset], [4, 4, 4], beg, end)
    iv = keys[:, 0]
    xv = keys[:, 1]
    ov = keys[:, 2]
    x = np.arange(beg, end)

    if figsize is None:
        figsize = (12, 4)
    plt.figure(figsize=figsize)
    plt.subplot(1, 3, 1)
    plt.plot(x, iv)
    plt.xlabel('Trace Number')
    plt.ylabel(f"Inline (at {iline})")
    plt.grid(True, linestyle='--')
    plt.subplot(1, 3, 2)
    plt.plot(x, xv)
    plt.xlabel('Trace Number')
    plt.ylabel(f"Crossline (at {xline})")
    plt.grid(True, linestyle='--')
    plt.subplot(1, 3, 3)
    plt.plot(x, ov)
    plt.xlabel('Trace Number')
    plt.ylabel(f"offset (at {offset})")
    plt.grid(True, linestyle='--')
    plt.tight_layout()
    if save:
        plt.savefig(save, dpi=200, bbox_inches='tight', pad_inches=0.0)
    plt.show()


def extract_arbitrary_line_by_view(data: np.ndarray,
                                   bmap: str = 'data',
                                   draw_arb: bool = True,
                                   *,
                                   line: bool = True,
                                   idx: int = 50,
                                   cline='#F3AA3C'):
    """
    extract arbitrary line from seismic data by clicking

    Parameters
    ----------
    - data: np.ndarray 
        3D seismic data
    - bmap: str
        background map, 'data' or 'blank'
    - line : bool
        whether to draw the broken line 
    - idx: int 
        the slice index of the seismic data if bmap is 'data'
    - cline: str
        color of the line

    Returns
    -------
    - out: np.ndarray
        extracted arbitrary line
    - p: np.ndarray
        extracted arbitrary line path
    - coords: np.ndarray
        the coordinates by clicking
    """
    fig, ax = plt.subplots()
    if bmap == 'data':
        img = data[:, :, idx]
        ax.imshow(img.T, cmap='gray', aspect='auto')
    else:
        img = np.zeros((data.shape[1], data.shape[0], 4), dtype=np.uint8)
        img = img + int(0.9 * 255)
        ax.imshow(img, aspect='auto')
        ax.grid(True)

    ax.set_title('Obtain a path by clicking, press "u" to undo, "enter" to finish') # yapf: disable
    ax.set_xlabel('Inline')
    ax.set_ylabel('Crossline')
    coords = []
    lines = []
    points = []
    out = None
    p = None
    indices = None

    def _draw_arbline():
        if (len(coords) > 0):
            ax.clear()
            nonlocal out, p, indices
            out, p, indices = arbitray_line(data, coords)
            ax.grid(False)
            ax.imshow(out.T, cmap='gray', aspect='auto')
            if line:
                for i in range(len(indices)):
                    ax.plot([indices[i]] * out.shape[1],
                            range(out.shape[1]),
                            'w--',
                            lw=0.5)
                    if i > 0 and i < len(indices) - 1:
                        ax.text(indices[i] - out.shape[0] / 50,
                                30,
                                str(indices[i]),
                                color='w',
                                fontsize=10)
            ax.set_title("Arbitrary Line")
            ax.set_ylabel('Time')
            fig.canvas.draw()
            # plt.close()

    def _click_event(event):
        if event.inaxes:
            coords.append((round(event.xdata, 2), round(event.ydata, 2)))
            p = ax.plot(event.xdata, event.ydata, 'ro')[0]
            points.append(p)
            if len(coords) > 1:
                x_coords, y_coords = zip(*coords[-2:])
                line, = ax.plot(x_coords, y_coords, c=cline)
                lines.append(line)
            fig.canvas.draw()

    def _undo_last(event):
        if event.key == 'u':
            if coords:
                coords.pop()
                p = points.pop()
                p.remove()
                if len(lines) > 0:
                    l = lines.pop()
                    l.remove()
                fig.canvas.draw()
        if event.key in ('enter', 'return', 'escape'):
            if draw_arb:
                fig.canvas.mpl_disconnect(cid_click)
                fig.canvas.mpl_disconnect(cid_key)
                _draw_arbline()
            else:
                plt.close(fig)

    cid_click = fig.canvas.mpl_connect('button_press_event', _click_event)
    cid_key = fig.canvas.mpl_connect('key_press_event', _undo_last)

    plt.show()

    return out, p, indices, np.array(coords)



# Copyright (c) 2024 Jintao Li.
# Computational and Interpretation Group (CIG),
# University of Science and Technology of China (USTC).
# All rights reserved.

import os, sys
import subprocess
from pathlib import Path
from setuptools import setup, find_packages
from distutils.ccompiler import get_default_compiler


def install_package(package):
    subprocess.check_call([sys.executable, "-m", "pip", "install", package])


try:
    from pybind11.setup_helpers import Pybind11Extension, build_ext
except ImportError:
    install_package("pybind11")
    from pybind11.setup_helpers import Pybind11Extension, build_ext

cwd = Path(__file__).resolve().parent

package_name = "cigsegy"

version_file = cwd / 'VERSION.txt'
with open(version_file) as vf:
    version = vf.read().strip()

if not version:
    raise RuntimeError("Failed to parse version from VERSION")

if get_default_compiler() == 'msvc':
    extra_compile_args = ['/wd4244', '/wd4996', '/wd4819']
else:
    extra_compile_args = ["-std=c++14", "-O3", "-undefined dynamic_lookup"]

extra_compile_args += ["-DUSE_PYBIND11"]

ext_modules = [
    Pybind11Extension(
        "cigsegy.cpp._CXX_SEGY",
        [
            "cigsegy/cpp/segywrap.cpp",
            "cigsegy/cpp/segyrw.cpp",
        ],
        extra_compile_args=extra_compile_args,
    ),
]

setup(
    name=package_name,
    version=version,
    long_description=open('README.rst').read(),
    long_description_content_type='text/x-rst',
    author='Jintao Li',
    url='https://github.com/JintaoLee-Roger/cigsegy',
    license='MIT',
    install_requires=['numpy', 'tqdm'],
    python_requires=">=3.6",
    setup_requires=['pybind11'],
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
    packages=find_packages(exclude=['docs', 'python', 'tools', 'tests']),
    include_package_data=True,
    package_data={
        "cigse": ["*.pyi"],
        "cigse.cpp": ["*.pyi"],
    },
)
